import pandas as pd
import pytest
from sklearn.utils.estimator_checks import check_estimator

from tc_uc4.datapreparation.prep import PreprocessingClass


test_with_data = [
    ('teleconsulto','prophet','data_richiesta','id_teleconsulto')
]
test_no_data = [
    ('teleconsulto','prophet')
]


@pytest.mark.parametrize("usecase,model,col1,col2", test_with_data)
def test_preprocessing_class_fit_transform(usecase,model,col1,col2):
  
    # Test the fit() and transform() methods of PreprocessingClass
    example_data = pd.DataFrame({col1: pd.to_datetime(['2023-06-01 12:00:00', '2023-06-02 14:30:00', '2023-06-03 15:25:00']), col2: ['A', 'B', 'C']})
    example_data_num2 = pd.DataFrame({col1: pd.to_datetime(['2023-06-01 12:00:00', '2023-06-02 12:40:00', '2023-06-02 12:00:00']), col2: ['A', 'B', 'C']})


    # Create an instance of PreprocessingClass
    preprocessing = PreprocessingClass(usecase=usecase, model=model)

    # Fit the PreprocessingClass on the example data
    result = preprocessing.fit_transform(X=example_data, time_granularity="D")

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame)
    # Check if the result has the same shape as the input data
    assert result.shape == example_data.shape

    # Fit the PreprocessingClass on the second example data
    result = preprocessing.fit_transform(X=example_data_num2, time_granularity="D")
    expected = pd.DataFrame({'Timestamp':  pd.to_datetime(['2023-06-01', '2023-06-02']), 'Target': [1.0,2.0]})

    # Check if the result has the same shape as the input data
    print(result)
    assert result.shape == (2,2)
    # Check if the data is equal to the expected dataframe
    pd.testing.assert_frame_equal(result, expected)

@pytest.mark.parametrize("usecase,model", test_no_data)
def test_preprocessing_class_save_load(tmp_path,usecase,model):
    # Test the save() and load() methods of PreprocessingClass

    # Create an instance of PreprocessingClass
    preprocessing = PreprocessingClass(usecase=usecase, model=model)

    # Save the PreprocessingClass instance to a file
    file_path = tmp_path / "preprocessing.pkl"
    preprocessing.save(file_path)

    # Load the PreprocessingClass instance from the file
    loaded_preprocessing = PreprocessingClass.load(file_path)

    # Check if the loaded instance is of the same type as the original instance
    assert isinstance(loaded_preprocessing, PreprocessingClass)

@pytest.mark.parametrize("usecase,model", test_no_data)
def test_preprocessing_class_estimator_compatibility(usecase,model):
    # Test the estimator compatibility of PreprocessingClass using sklearn's check_estimator

    # Create an instance of PreprocessingClass
    preprocessing = PreprocessingClass(usecase=usecase, model=model)

    # Check if PreprocessingClass meets the requirements of an estimator
    assert check_estimator(preprocessing, generate_only=True)



import pandas as pd
import numpy as np
import pytest
from tc_uc4.datapreparation.datapreparation_utils import (PreprocessingTeleconsulto, code_to_name, Normalizer,
                                                          ReplacerNA, Detrender, Deseasoner, Differencer)
from tc_uc4.utility.exceptions import InputTypeError
from statsmodels.tsa.seasonal import seasonal_decompose


### PreprocessingTeleconsulto ###

@pytest.fixture
def sample_df():
    data = {
        'data_richiesta': pd.date_range(start='2022-01-01', end='2022-01-10'),
        'id_teleconsulto': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    }
    return pd.DataFrame(data)


def test_prophet(sample_df):
    preprocessing = PreprocessingTeleconsulto()

    # Test with default time_granularity
    expected_result = pd.DataFrame({
        'Timestamp': pd.date_range(start='2022-01-01', end='2022-01-10'),
        'Target': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    })
    result = preprocessing.prophet(sample_df, time_granularity="D")
    assert result.equals(expected_result)

    # Test with custom time_granularity
    expected_result = pd.DataFrame({
        'Timestamp': pd.date_range(start='2022-01-01', end='2022-01-10', freq='12H'),
        'Target': [1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0, np.nan, 1.0]
    })
    result = preprocessing.prophet(sample_df, time_granularity="12H")
    assert result.equals(expected_result)

    # Test with invalid input type
    with pytest.raises(InputTypeError):
        preprocessing.prophet("invalid_dataframe")

    # Test with invalid datetime_index
    with pytest.raises(InputTypeError):
        preprocessing.prophet(sample_df.rename(columns={'data_richiesta': 'invalid_index'}))

    # Test with invalid target
    with pytest.raises(InputTypeError):
        preprocessing.prophet(sample_df.rename(columns={'id_teleconsulto': 'invalid_target'}))


### code_to_name ###

# Test case for valid input
def test_code_to_name_valid_input():
    # Create a sample pd.Series and mapping dictionary for testing
    cod = pd.Series([1, 2, 3, 4, 5])
    convers_dict = {1: "Apple", 2: "Banana", 3: "Orange", 4: "Grape", 5: "Mango"}
    # Call the code_to_name function
    result = code_to_name(cod, convers_dict)
    assert isinstance(result, pd.Series) #the result should be a Series

    expected_result = pd.Series(["Apple", "Banana", "Orange", "Grape", "Mango"])
    pd.testing.assert_series_equal(result, expected_result) #the result should has the correct value

# Test case for invalid input
def test_code_to_name_invalid_input():
    # Create an invalid input (not a pd.Series)
    invalid_cod = [1, 2, 3, 4, 5]
    convers_dict = {1: "Apple", 2: "Banana", 3: "Orange", 4: "Grape", 5: "Mango"}
    # Test for InputTypeError when invalid input is provided for cod
    with pytest.raises(InputTypeError):
        code_to_name(invalid_cod, convers_dict)

    # Create an invalid input (not a dict)
    cod = pd.Series([1, 2, 3, 4, 5])
    invalid_convers_dict = [1, 2, 3, 4, 5]
    # Test for InputTypeError when invalid input is provided for convers_dict
    with pytest.raises(InputTypeError):
        code_to_name(cod, invalid_convers_dict)


### Normalizer ###

@pytest.fixture
def example_data():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_fit_normalizer(example_data):
    # Create an instance of the Normalizer class
    normalizer = Normalizer()

    # Call the fit method with the example data
    fitted_normalizer = normalizer.fit(example_data)

    # Verify that the output is the same instance of normalizer
    assert fitted_normalizer is normalizer

    # Verify that the min and max values are calculated correctly
    assert normalizer.min == 10
    assert normalizer.max == 50

def test_transform_normalizer(example_data):
    # Create an instance of the Normalizer class
    normalizer = Normalizer()

    # Call the fit method with the example data
    normalizer.fit(example_data)

    # Call the transform method with the example data
    transformed_data = normalizer.transform(example_data)
    print(transformed_data)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the DataFrame has the correct columns
    assert transformed_data.columns.tolist() == ['timestamp', 'volumes']

    # Verify that the data is correctly normalized
    assert transformed_data['volumes'].min() == 0
    assert transformed_data['volumes'].max() == 1

    # Verify that the data has not been modified in other ways
    assert example_data['timestamp'].equals(transformed_data['timestamp'])


### ReplacerNA ###

#transform function
@pytest.fixture
def example_data_transform():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, None, 30, None, 50]})
    return data

def test_fit_replace_na(example_data_transform):
    # Create an instance of the ReplacerNA class with method="mean"
    replacer = ReplacerNA(method="mean")

    # Call the fit method with the example data
    fitted_replacer = replacer.fit(example_data_transform)

    # Verify that the output is the same instance of replacer
    assert fitted_replacer is replacer

    # Verify that the value and method_for_df are calculated correctly
    assert replacer.value == 30.0
    assert replacer.method_for_df is None

def test_transform_replace_na(example_data_transform):
    # Create an instance of the ReplacerNA class with method="zero"
    replacer = ReplacerNA(method="zero")

    # Call the fit method with the example data
    replacer.fit(example_data_transform)

    # Call the transform method with the example data
    transformed_data = replacer.transform(example_data_transform)
    print(transformed_data)
    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data has missing values replaced correctly
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': [10.0, 0.0, 30.0, 0.0, 50.0]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)

def test_transform__replace_na_ffill(example_data_transform):
    # Create an instance of the ReplacerNA class with method="ffill"
    replacer = ReplacerNA(method="ffill")

    # Call the fit method with the example data
    replacer.fit(example_data_transform)

    # Call the transform method with the example data
    transformed_data = replacer.transform(example_data_transform)

    # Verify that the transformed data has missing values replaced correctly using forward fill
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': [10.0, 10.0, 30.0, 30.0, 50.0]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Detrender ###

#trend function
@pytest.fixture
def example_data_trend():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_fit_detrender(example_data_trend):
    # Create an instance of the Detrender class with period=2
    detrender = Detrender(period=2)

    # Call the fit method with the example data
    fitted_detrender = detrender.fit(example_data_trend)

    # Verify that the output is the same instance of detrender
    assert fitted_detrender is detrender

    # Verify that the trend is calculated correctly
    expected_trend = pd.Series([10.0, 20.0, 30.0, 40.0, 50.0], name='trend')
    pd.testing.assert_series_equal(detrender.trend, expected_trend)

def test_transform_detrender(example_data_trend):
    # Create an instance of the Detrender class with period=2
    detrender = Detrender(period=2)

    # Call the fit method with the example data
    detrender.fit(example_data_trend)

    # Call the transform method with the example data
    transformed_data = detrender.transform(example_data_trend)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is detrended correctly
    expected_trend = seasonal_decompose(example_data_trend['volumes'], model='additive', period=2, extrapolate_trend='freq').trend
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': example_data_trend['volumes'] - expected_trend})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Deseasoner ###

#season function
@pytest.fixture
def example_data_season():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_transform_deseasoner(example_data_season):
    # Create an instance of the Deseasoner class with period=2
    deseasoner = Deseasoner(period=2)

    # Call the fit method with the example data
    deseasoner.fit(example_data_season)

    # Call the transform method with the example data
    transformed_data = deseasoner.transform(example_data_season)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is deseasonalized correctly
    expected_seasonal = seasonal_decompose(example_data_season['volumes'], model='additive', period=2, extrapolate_trend='freq').seasonal
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': example_data_season['volumes'] - expected_seasonal})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Differencer ###

#difference function
@pytest.fixture
def example_data_difference():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_transform_differencer(example_data_difference):
    # Create an instance of the Differencer class with lag=1
    differencer = Differencer(lag=1)

    # Call the fit method with the example data
    differencer.fit(example_data_difference)

    # Call the transform method with the example data
    transformed_data = differencer.transform(example_data_difference)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is differenced correctly
    expected_data = pd.DataFrame({'timestamp': [2, 3, 4, 5],
                                  'volumes': [10, 10, 10, 10]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)



import pandas as pd
import pytest
import os
from tc_uc4.algorithms.prophet_utils import (check_prophet_fit_df, check_prophet_predict_df,
                                      check_prophet_predict_df, check_prophet_save,
                                      check_prophet_load, check_split_input, train_val_test_split,
                                      save_model_results, preprocess_prophet_output,
                                      check_preprocess_prophet_input, preprocess_prophet_input,
                                      is_not_convertible_to_int_float, literal_evaluation, generate_values,
                                      grid_values_hyperparameters)

### check_prophet_fit_df ###

# Test case for a valid input DataFrame
def test_check_prophet_fit_df_valid_input():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')],
                              'y': [10, 20]})
    assert check_prophet_fit_df(input_data) is None

# Test case for invalid input (not a DataFrame)
def test_check_prophet_fit_df_invalid_input():
    input_data = 'invalid'
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

def test_check_prophet_fit_df_empty_dataframe():
    input_data = pd.DataFrame({'ds': []})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

# Test case for DataFrame with less than two columns
def test_check_prophet_fit_df_less_than_two_columns():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')]})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

# Test case for DataFrame missing 'ds' column
def test_check_prophet_fit_df_missing_ds_column():
    input_data = pd.DataFrame({'y': [10, 20]})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

# Test case for DataFrame missing 'y' column
def test_check_prophet_fit_df_missing_y_column():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')]})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

# Test case for non-datetime 'ds' column
def test_check_prophet_fit_df_non_datetime_ds_column():
    input_data = pd.DataFrame({'ds': ['2023-01-01', '2023-01-02'],
                              'y': [10, 20]})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)

# Test case for non-numeric 'y' column
def test_check_prophet_fit_df_non_numeric_y_column():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')],
                              'y': ['10', '20']})
    with pytest.raises(ValueError):
        check_prophet_fit_df(input_data)



### check_prophet_predict_df ###

# Test case for a valid input DataFrame
def test_check_prophet_predict_df_valid_input():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')]})
    assert check_prophet_predict_df(input_data) is None

# Test case for invalid input (not a DataFrame)
def test_check_prophet_predict_df_invalid_input():
    input_data = 'invalid'
    with pytest.raises(ValueError):
        check_prophet_predict_df(input_data)

# Test case for DataFrame with no columns
def test_check_prophet_predict_df_no_columns():
    input_data = pd.DataFrame()
    with pytest.raises(ValueError):
        check_prophet_predict_df(input_data)

# Test case for an empty DataFrame
def test_check_prophet_predict_df_empty_dataframe():
    input_data = pd.DataFrame({'ds': []})
    with pytest.raises(ValueError):
        check_prophet_predict_df(input_data)

# Test case for DataFrame missing 'ds' column
def test_check_prophet_predict_df_missing_ds_column():
    input_data = pd.DataFrame({'y': [10, 20]})
    with pytest.raises(ValueError):
        check_prophet_predict_df(input_data)

# Test case for non-datetime 'ds' column
def test_check_prophet_predict_df_non_datetime_ds_column():
    input_data = pd.DataFrame({'ds': [pd.Timestamp('2023-01-01'), pd.Timestamp('2023-01-02')]})
    input_data['ds'] = pd.to_numeric(input_data['ds'])
    with pytest.raises(ValueError):
        check_prophet_predict_df(input_data)


### check_prophet_save ###

class DummyModel:
    def __init__(self, history):
        if history:
            history = pd.DataFrame()
        else:
            history = 'Not valid history'

        self.history = history

# Test case for a valid model and file path
def test_check_prophet_save_valid_input(tmpdir):
    model = DummyModel(history=True)
    file_path = os.path.join(tmpdir, 'model.pkl')
    assert check_prophet_save(model, file_path) is None

# Test case for invalid file extension
def test_check_prophet_save_invalid_file_extension(tmpdir):
    model = DummyModel(history=True)
    file_path = os.path.join(tmpdir, 'model.txt')
    with pytest.raises(ValueError):
        check_prophet_save(model, file_path)

# Test case for saving an untrained model
def test_check_prophet_save_untrained_model(tmpdir):
    model = DummyModel(history=False)
    file_path = os.path.join(tmpdir, 'model.pkl')
    with pytest.raises(ValueError):
        check_prophet_save(model, file_path)

# Test case for non-existent directory
def test_check_prophet_save_nonexistent_directory(tmpdir):
    model = DummyModel(history=True)
    file_path = os.path.join(tmpdir, 'subdir/model.pkl')
    with pytest.raises(ValueError):
        check_prophet_save(model, file_path)


### check_prophet_load ###

# Test case for a valid file path
def test_check_prophet_load_valid_input(tmpdir):
    file_path = os.path.join(tmpdir, 'model.pkl')
    open(file_path, 'a').close()  # Create an empty file
    assert check_prophet_load(file_path) is None

# Test case for non-existent file path
def test_check_prophet_load_nonexistent_file(tmpdir):
    file_path = os.path.join(tmpdir, 'nonexistent.pkl')
    with pytest.raises(ValueError):
        check_prophet_load(file_path)

# Test case for invalid file extension
def test_check_prophet_load_invalid_file_extension(tmpdir):
    file_path = os.path.join(tmpdir, 'model.txt')
    open(file_path, 'a').close()  # Create an empty file
    with pytest.raises(ValueError):
        check_prophet_load(file_path)


### check_split_input ###

# Test case for a valid input DataFrame and validation size
def test_check_split_input_valid_input():
    df = pd.DataFrame({'datetime': pd.date_range(start='2022-01-01', periods=5)})
    val_size = 0.2
    assert check_split_input(df, val_size) is None

# Test case for an empty DataFrame
def test_check_split_input_empty_dataframe():
    df = pd.DataFrame()
    val_size = 0.2
    with pytest.raises(ValueError):
        check_split_input(df, val_size)

# Test case for DataFrame missing datetime column
def test_check_split_input_missing_datetime_column():
    df = pd.DataFrame({'col1': [1, 2, 3]})
    val_size = 0.2
    with pytest.raises(ValueError):
        check_split_input(df, val_size)

# Test case for non-monotonic datetime column
def test_check_split_input_non_monotonic_datetime_column():
    df =pd.DataFrame({'datetime': [pd.Timestamp('2022-01-01'), pd.Timestamp('2021-01-03'), pd.Timestamp('2021-01-05'),
                                    pd.Timestamp('2021-01-03'), pd.Timestamp('2021-01-01')]})
    val_size = 0.2
    with pytest.raises(ValueError):
        check_split_input(df, val_size)

# Test case for invalid validation size
def test_check_split_input_invalid_val_size():
    df = pd.DataFrame({'datetime': pd.date_range(start='2022-01-01', periods=5)})
    val_size = -0.2
    with pytest.raises(ValueError):
        check_split_input(df, val_size)

# Test case for invalid test size
def test_check_split_input_invalid_test_size():
    df = pd.DataFrame({'datetime': pd.date_range(start='2022-01-01', periods=5)})
    val_size = 0.2
    test_size = 1.5
    with pytest.raises(ValueError):
        check_split_input(df, val_size, test_size)


### train_val_test_split ###

@pytest.fixture
def example_df_split():
    return pd.DataFrame({'datetime': pd.to_datetime(['2022-01-01', '2022-01-02', '2022-01-03', '2022-01-04', '2022-01-05']), 'col1': [1, 2, 3, 4, 5]})

# Test case for a valid input DataFrame and validation size
def test_train_val_test_split_valid_input(example_df_split):
    # df = pd.DataFrame({'datetime': [pd.Timestamp('2022-01-01'), pd.Timestamp('2022-01-02'), pd.Timestamp('2022-01-03'),
    #                                 pd.Timestamp('2022-01-04'), pd.Timestamp('2022-01-05')], 'col1': [1, 2, 3, 4, 5]})
    val_size = 0.2
    test_size = 0.1
    train_set, val_set, test_set = train_val_test_split(example_df_split, val_size, test_size)
    
    assert len(train_set) == 3
    assert len(val_set) == 1
    assert len(test_set) == 1

# Test case for a valid input DataFrame and validation size only
def test_train_val_test_split_valid_input_without_test_size(example_df_split):
    # df = pd.DataFrame({'datetime': [pd.Timestamp('2022-01-01'), pd.Timestamp('2022-01-02'), pd.Timestamp('2022-01-03'),
    #                                 pd.Timestamp('2022-01-04'), pd.Timestamp('2022-01-05')],'col1': [1, 2, 3, 4, 5]})
    val_size = 0.2
    train_set, val_set = train_val_test_split(example_df_split, val_size)
    
    assert len(train_set) == 4
    assert len(val_set) == 1

# Test case for invalid validation size
def test_train_val_test_split_invalid_val_size(example_df_split):
    # df = pd.DataFrame({'datetime': [pd.Timestamp('2022-01-01'), pd.Timestamp('2022-01-02'), pd.Timestamp('2022-01-03'),
    #                                 pd.Timestamp('2022-01-04'), pd.Timestamp('2022-01-05')],'col1': [1, 2, 3, 4, 5]})
    val_size = -0.2
    test_size = 0.1
    with pytest.raises(ValueError):
        train_val_test_split(example_df_split, val_size, test_size)

# Test case for invalid test size
def test_train_val_test_split_invalid_test_size(example_df_split):
    # df = pd.DataFrame({'datetime': [pd.Timestamp('2022-01-01'), pd.Timestamp('2022-01-02'), pd.Timestamp('2022-01-03'),
    #                                 pd.Timestamp('2022-01-04'), pd.Timestamp('2022-01-05')],'col1': [1, 2, 3, 4, 5]})
    val_size = 0.2
    test_size = 1.5
    with pytest.raises(ValueError):
        train_val_test_split(example_df_split, val_size, test_size)


### save_model_results, preprocess_prophet_output ###

@pytest.fixture
def example_df_output():
    return pd.DataFrame({
        'ds':  pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'yhat': [10.5, 11.2, 9.8],
        'yhat_lower': [9.5, 10.8, 9.0],
        'yhat_upper': [11.5, 11.6, 10.5]
    })

def test_preprocess_prophet_output(example_df_output):
    id_pred = 'Lombardia'
    expected_output = pd.DataFrame({
        'Timestamp': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'Id_pred': ['Lombardia', 'Lombardia', 'Lombardia'],
        'Pred_mean': [10, 11, 9],
        'Sigma': [1, 0, 0],
        'Pi_lower_95': [9, 10, 9],
        'Pi_upper_95': [11, 11, 10]
    })

    output = preprocess_prophet_output(example_df_output, id_pred)
    pd.testing.assert_frame_equal(output, expected_output)

def test_save_model_results(example_df_output, tmp_path):
    path = str(tmp_path)
    id_pred = 'Lombardia'
    save_model_results(example_df_output, path, id_pred, 'TEST.parquet')

    assert 'TEST.parquet' in os.listdir(str(tmp_path))


### check_preprocess_prophet_input, preprocess_prophet_input ###

@pytest.fixture
def example_df_input():
    return pd.DataFrame({
        'Timestamp': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'Target': [10.5, 11.2, 9.8]
    })

def test_check_preprocess_prophet_input_invalid_dataframe():
    with pytest.raises(ValueError, match="Input is not a DataFrame."):
        check_preprocess_prophet_input('not_a_dataframe', 'Timestamp', 'Target')

def test_check_preprocess_prophet_input_invalid_empty_dataframe():
    empty_df = pd.DataFrame()
    with pytest.raises(ValueError, match="Il DataFrame è vuoto."):
        check_preprocess_prophet_input(empty_df, 'Timestamp', 'Target')

def test_check_preprocess_prophet_input_invalid_less_than_two_columns():
    df = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-01'])})
    with pytest.raises(ValueError, match="DataFrame must have at least two columns."):
        check_preprocess_prophet_input(df, 'Timestamp', 'Target')

def test_check_preprocess_prophet_input_invalid_missing_columns():
    df = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']), 'z': [10.5, 11.2, 9.8]})
    with pytest.raises(ValueError, match="Specified columns are not present in the DataFrame."):
        check_preprocess_prophet_input(df, 'Timestamp', 'Target')

def test_check_preprocess_prophet_input_invalid_inconsistent_columns():
    df = pd.DataFrame({'Timestamp': [1, 2, 3], 'Target': [10.5, 11.2, 9.8]})
    with pytest.raises(ValueError, match="The column Timestamp must have datetime type"):
        check_preprocess_prophet_input(df, 'Timestamp', 'Target')
    
    df = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']), 'Target': ['A', 'B', 'C']})
    with pytest.raises(ValueError, match="The column Target must have numeric type"):
        check_preprocess_prophet_input(df, 'Timestamp', 'Target')

def test_preprocess_prophet_input(example_df_input, capsys):
    processed_df = preprocess_prophet_input(example_df_input, 'Timestamp', 'Target')
    captured = capsys.readouterr()
    assert captured.out == ""

    expected_df = pd.DataFrame({
        'ds': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'y': [10.5, 11.2, 9.8]
    })
    pd.testing.assert_frame_equal(processed_df, expected_df)

def test_preprocess_prophet_input_invalid(example_df_input):
    with pytest.raises(ValueError, match="Specified columns are not present in the DataFrame."):
        preprocess_prophet_input(example_df_input, 'invalid_date', 'invalid_target')

    processed_df = preprocess_prophet_input(example_df_input, 'Timestamp', 'Target')
    expected_df = pd.DataFrame({
        'ds': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'y': [10.5, 11.2, 9.8]
    })
    pd.testing.assert_frame_equal(processed_df, expected_df)


### is_not_convertible_to_int_float ###

def test_is_not_convertible_to_int_float():
    assert is_not_convertible_to_int_float("abc") == True
    assert is_not_convertible_to_int_float("1.23.4") == True
    assert is_not_convertible_to_int_float("123") == False
    assert is_not_convertible_to_int_float("1.23") == False


### literal_evaluation ###

def test_literal_evaluation():
    values_list = ['True', 'False', '123', '12.3']
    assert literal_evaluation(values_list) == [True, False, 123, 12.3]

    values_list = ['True', 'False', '1.23', 'abc']
    with pytest.raises(ValueError):
        literal_evaluation(values_list)


### generate_values ###

def test_generate_values():
    dictionary = {'min': 0, 'max': 5, 'step': 1}
    assert generate_values(dictionary) == [0, 1, 2, 3, 4, 5]

    dictionary = {'min': 0, 'max': 5, 'step': 0.5}
    assert generate_values(dictionary) == [0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5]

    # min is 0 so i get the default_step
    dictionary = {'min': 0, 'max': 5, 'step': -1}
    assert generate_values(dictionary) == [0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5]
    # min is not 0 so i get the min as step
    dictionary = {'min': 1, 'max': 5, 'step': -1}
    assert generate_values(dictionary) == [1, 2, 3, 4, 5]

    dictionary = {'min': 0, 'max': 5, 'step': 10}
    with pytest.raises(ValueError):
        generate_values(dictionary)


### grid_values_hyperparameters ###

def test_grid_values_hyperparameters():
    config_toml = {
        'param1': ['True', 'False'],
        'changepoint_prior_scale': {'min': 0, 'max': 5, 'step': 1},
        'seasonality_prior_scale': {'min': 0, 'max': 5, 'step': 0.5}
    }
    expected_result = {
        'param1': [True, False],
        'changepoint_prior_scale': [0, 1, 2, 3, 4, 5],
        'seasonality_prior_scale': [0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5]
    }
    assert grid_values_hyperparameters(config_toml) == expected_result

    config_toml = {
        'param1': ['True', 'False'],
        'changepoint_prior_scale': {'min': 0, 'max': 5, 'step': 10},
        'seasonality_prior_scale': {'min': 0, 'max': 5, 'step': 0.1}
    }
    with pytest.raises(ValueError):
        grid_values_hyperparameters(config_toml)



import pandas as pd
import numpy as np
import os
import ast

from ..datahandler.datahandler import DataHandler
from ..utility.tele_logger import logger

def check_prophet_fit_df(input_data):
    """
    Esegue i controlli sull'input fornito per il metodo fit() di Prophet.

    Parameters
    ----------
    input_data
        L'input fornito per il metodo fit() di Prophet.

    Raises
    ------
    ValueError
        Se l'input non è del tipo pd.DataFrame.
    ValueError
        Se il dataframe non ha almeno due colonne.
    ValueError
        Se il dataframe non contiene le colonne 'ds' e 'y'.
    ValueError
        Se la colonna 'ds' non è di tipo datetime.
    ValueError
        Se la colonna 'y' non è di tipo numerico
    """

    # Controllo se l'input è un DataFrame
    if not isinstance(input_data, pd.DataFrame):
        raise ValueError("L'input deve essere un DataFrame.")

    # Controllo se il DataFrame ha almeno due colonne
    if len(input_data.columns) < 2:
        raise ValueError("Il DataFrame deve avere almeno due colonne.")

    if len(input_data) == 0:
        raise ValueError("Il DataFrame è vuoto.")
    
    # Controllo se 'ds' e 'y' sono presenti nel DataFrame
    if 'ds' not in input_data.columns or 'y' not in input_data.columns:
        raise ValueError("Il DataFrame deve contenere le colonne 'ds' e 'y'.")

    # Controllo il tipo delle colonne
    if not pd.core.dtypes.common.is_datetime64_any_dtype(input_data['ds']):
        raise ValueError("La colonna 'ds' deve essere di tipo datetime.")

    if not pd.core.dtypes.common.is_numeric_dtype(input_data['y']):
        raise ValueError("La colonna 'y' deve essere di tipo numerico.")

def check_prophet_predict_df(input_data):
    """
    Esegue i controlli sull'input fornito per il metodo predict() di Prophet.

    Parameters
    ----------
    input_data
        L'input fornito per il metodo predict() di Prophet.

    Raises
    ------
    ValueError
        Se l'input fornito non è del tipo pd.DataFrame
    ValueError
        Se il dataframe non ha almeno una colonna
    ValueError
        Se la colonna 'ds' non è presente nel dataframe
    ValueError
        Se la colonna 'ds' non è del tipo datetime
    """
    # Controllo se l'input è un DataFrame
    if not isinstance(input_data, pd.DataFrame):
        raise ValueError("L'input deve essere un DataFrame.")

    # Controllo se il DataFrame ha almeno una colonna
    if len(input_data.columns) < 1:
        raise ValueError("Il DataFrame deve avere almeno una colonna.")

    if len(input_data) == 0:
        raise ValueError("Il DataFrame è vuoto.")
    
    # Controllo se 'ds' è presente nel DataFrame
    if 'ds' not in input_data.columns:
        raise ValueError("Il DataFrame deve contenere la colonna 'ds'.")
    # Controllo sul tipo della colonna ds
    if not pd.core.dtypes.common.is_datetime64_any_dtype(input_data['ds']):
        raise ValueError("La colonna 'ds' deve essere di tipo datetime.")
    
def check_prophet_save(model, file_path):
    """
    Esegue i controlli sull'input fornito per il metodo save() di Prophet.

    Parameters
    ----------
    model
        fitted prophet model
    file_path: str
        The complete file path where to save the model

    Raises
    ------
    ValueError
        Se il file non ha un' estensione pkl
    ValueError
        Se provi a salvare un modello non addestrato
    ValueError
        Se la cartella di destinazione per il salvataggio non esiste.
    """


    # Controllo sull'estensione del file
    if not file_path.endswith(".pkl"):
        raise ValueError("Il file deve avere un'estensione '.pkl' per il salvataggio del modello.")

    # Controllo se il modello è stato addestrato correttamente
    if not isinstance(model.history, pd.DataFrame):
        raise ValueError("Il modello non è stato addestrato correttamente. Esegui il metodo fit() prima del salvataggio.")

    # Controllo se il percorso del file esiste
    file_dir = os.path.dirname(file_path)
    if file_dir and not os.path.exists(file_dir):
        raise ValueError("La cartella di destinazione per il salvataggio non esiste.")

def check_prophet_load(file_path):
    """
    Esegue i controlli sull'input fornito per il metodo load() di Prophet.

    Parameters
    ----------
    file_path
        Il percorso del file contenente il modello da caricare.
    Raises
    ------
    ValueError
        Se il percorso del file non esiste.
    ValueError
        Se il file che provi a caricare non ha estensione pkl
    """

    # Controllo se il percorso del file esiste
    if not os.path.exists(file_path):
        raise ValueError("Il percorso del file non esiste.")

    # Controllo sull'estensione del file
    if not file_path.endswith(".pkl"):
        raise ValueError("Il file deve avere un'estensione '.pkl' per il caricamento del modello.")

def check_split_input(df, val_size, test_size=None):
    """
    Check the input parameters for split train, validation, test

    Parameters
    ----------
    df: pd.DataFrame
        The dataframe that needs to be split
    val_size:
        Percentage of validation set
    test_size:
        Percentage of test set (optional)
    Raises
    ------
    ValueError
        If the DataFrame is empty.
    ValueError
        If the dataframe does not contain a datetime column.
    ValueError
        If the datetime column is not sorted.
    ValueError
        If val_size is not between 0 and 1.
    ValueError
        If test_size is provided and not between 0 and 1.
    """
    if len(df) == 0:
        raise ValueError("Il DataFrame è vuoto.")

    # Controllo se esiste una colonna di tipo datetime
    datetime_columns = [col for col in df.columns if pd.api.types.is_datetime64_any_dtype(df[col])]
    if len(datetime_columns) == 0:
        raise ValueError("Il dataframe non contiene una colonna di tipo datetime")

    # Controllo se la colonna temporale è ordinata in modo monotono
    for col in datetime_columns:
        if not df[col].is_monotonic_increasing and not df[col].is_monotonic_decreasing:
            raise ValueError(f"La colonna '{col}' non è ordinata in modo monotono")

    if not (val_size > 0):
        raise ValueError("val_size must be greater or equal than 1")

    if test_size is not None and not (test_size > 0):
        raise ValueError("test_size must be greater or equal than 1")

def train_val_test_split(df: pd.DataFrame, val_size: float, test_size: float = None) -> tuple:
    """
    Create train, validation, test dataframes

    Parameters
    ----------
    df: pd.DataFrame
        The dataframe that needs to be split
    train_size: float, optional
        Percentage of training set
    val_size: float, optional
        Percentage of validation set
    test_size: float, optional
        Percentage of test set

    Returns
    -------
        Tuple: (train_set, val_set, test_set)
    """
    # Parameters check
    check_split_input(df = df, val_size = val_size, test_size = test_size)

    total_rows = len(df)

    # Effettua lo split dei dati
    train_set = df.iloc[:-(val_size+test_size)]
    val_set = df.iloc[(total_rows-val_size-test_size):(-1)*test_size]
    test_set = df.iloc[(total_rows-test_size):]

    return train_set, val_set, test_set
    
def preprocess_prophet_output(df: pd.DataFrame, id_pred: str) -> pd.DataFrame:
    """
    The function generates the results table from the Prophet table.
    The results table has columns (Timestamp, Id_pred, Pred_mean, Sigma, Pi_lower_95, Pi_upper_95)

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe representing the Prophet results
    id_pred: str
        String referring to the id_pred (e.g. Lombardia, Lombardia/ASL*)

    Returns
    -------
    pd.DataFrame
        The dataframe in the format (Timestamp, Id_pred, Pred_mean, Sigma, Pi_lower_95, Pi_upper_95)
    """
    preproc_df = df[['ds', 'yhat', 'yhat_lower', 'yhat_upper']].copy()
    preproc_df['Id_pred'] = id_pred
    preproc_df.columns = ['Timestamp', 'Pred_mean', 'Pi_lower_95', 'Pi_upper_95', 'Id_pred']
    preproc_df['Sigma'] = (preproc_df['Pi_upper_95'] - preproc_df['Pi_lower_95']) / (2 * 1.96)
    preproc_df[['Pred_mean', 'Pi_lower_95', 'Pi_upper_95', 'Sigma']] = preproc_df[['Pred_mean', 'Pi_lower_95', 'Pi_upper_95', 'Sigma']].astype(int)
    return preproc_df[['Timestamp', 'Id_pred', 'Pred_mean', 'Sigma', 'Pi_lower_95', 'Pi_upper_95']]

def save_model_results(df: pd.DataFrame, path_to_folder: str, id_pred: str, filename: str = 'output.parquet') -> None:
    """
    Method to save the prophet dataframe

    Parameters
    ----------
    df: pd.DataFrame
    Prophet dataframe that needs to be changed for saving

    path_to_folder: str
        path where to the folder where save the output '.parquet'
    
    filename: str, optional
        filename of the '.parquet' file

    id_pred: str
    String referring to the id_pred (e.g. Lombardia, Lombardia/ASL*)

    Returns
    -------
        None
"""
    DH = DataHandler(path_to_folder)
    try:
        result_df = preprocess_prophet_output(df, id_pred)
    except Exception as e:
        raise ValueError(f"Error occurred during the generation of the output file: {str(e)}. Make sure the dataset is the one resulted from Prophet predict.")
    
    logger.info(' Saving the model result')
    DH.write(result_df, folder='', filename = filename)

def check_preprocess_prophet_input(df, date, target):
    """
    Check the consistency of the dataframe
    Parameters
    ----------
    df
        input dataframe
    date
        date column name to check
    target
        target column name to check

    Returns
    -------
        the two columns ordered by type (datetime, numeric)

    Raises
    ------
    ValueError
        if df is not a dataframe
    ValueError
        if the dataframe has less than two columns
    ValueError
        if col1 and col2 are not dataframe columns
    ValueError
        if the columns are not datetime and numeric
    """
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Input is not a DataFrame.")
    
    if len(df) == 0:
        raise ValueError("Il DataFrame è vuoto.")

    if len(df.columns) < 2:
        raise ValueError("DataFrame must have at least two columns.")

    if date not in df.columns or target not in df.columns:
        raise ValueError("Specified columns are not present in the DataFrame.")

    if not pd.api.types.is_datetime64_any_dtype(df[date]):
        raise ValueError(f"The column {date} must have datetime type")
        
    if not pd.api.types.is_numeric_dtype(df[target]):
        raise ValueError(f"The column {target} must have numeric type")

def preprocess_prophet_input(df: pd.DataFrame, date: str, target: str):
    """
    Convert the date column name into 'ds' and the target column name in 'y'

    Parameters
    ----------
    df: pd.DataFrame
        Input DataFrame to check
    date: str
        Name of the date column
    target: str
        Name of the target column

    Returns
    -------
        the processed df
    """
    
    check_preprocess_prophet_input(df = df, date = date, target = target)
    df_ = df.rename(columns={date: 'ds', target: 'y'})
    return df_

def is_not_convertible_to_int_float(string_: str):
    """
    check if the string cannot be converted into float or int

    Parameters
    ----------
    string_
        the string to be parsed

    Returns
    -------
        False if the string cannot be converted, True otherwise
    """
    try:
        ast.literal_eval(string_)
        return False
    except ValueError:
        return True
    except SyntaxError:
        return True
    
def literal_evaluation(values_list: list) -> list:
    """
    The function evaluates the list values for the parameters: 'weekly_seasonality', 'yearly_seasonality'

    Parameters
    ----------
    values_list
        list of strings to be evaluates

    Returns
    -------
        list of evaluated values
    """
    for j in values_list:
        if j not in ['True', 'False'] and is_not_convertible_to_int_float(j):
            raise ValueError(f'The following value {j} is not a valid parameter')
    return [ast.literal_eval(val) for val in values_list]

def generate_values(dictionary, default_step = 0.5):
    """
    The function generates the possible values of the numeric prophet parameters

    Parameters
    ----------
    dictionary
        dictionary formt he config file with min, max and step
    default_step, optional
        the default step used if it's not specified in the config file

    Returns
    -------
        list of possible values

    Raises
    ------
    ValueError
        if the step is greater than the values range
    """

    min_value = dictionary.get('min')
    max_value = dictionary.get('max')
    step = dictionary.get('step')
    if step == 'None':
        step = default_step  # Valore di default dello step

    if not isinstance(min_value, (float, int)) or not isinstance(max_value, (float, int)) or not isinstance(step, (float, int)):
        raise ValueError(f' One of the dtypes of {min_value}, {max_value}, {step} is not valid')
    
    if step == -1 and min_value>0:
        step = min_value  # Utilizza min value come step
    elif step == -1 and min_value==0:
        step = default_step
    elif step > (max_value - min_value):
        raise ValueError(f'The step {step} must be lower than the difference between max {max_value} and min {min_value}')

    values = list(np.arange(min_value,max_value,step))

    # Verifica se il valore massimo è già incluso nella lista
    if max_value not in values:
        values.append(max_value)

    return values

def grid_values_hyperparameters(config_toml: dict) -> dict:
    """
    The function takes as input the dictionary of the config file
    and generates the combination of grid search parameters

    Parameters
    ----------
    config_toml
        the dictionary form the config file

    Returns
    -------
        the grid search parameter dictionary
    """
    grid_parameters = {}
    for param in config_toml.keys():
        if (type(config_toml[param]) == list):
            grid_parameters[param] = literal_evaluation(config_toml[param])

        elif (type(config_toml[param]) == dict):
            if param.lower() == 'changepoint_prior_scale':
                grid_parameters[param] = generate_values(config_toml[param], default_step = 0.1)

            if param.lower() == 'seasonality_prior_scale':
                grid_parameters[param] = generate_values(config_toml[param], default_step = 1.0)
        else:
            raise ValueError(f' The value datetype of {param} is not valid')
    return grid_parameters


import pandas as pd
import numpy as np
from statsmodels.tsa.seasonal import seasonal_decompose
from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self
from typing import Union, Dict, List

from ..utility.tele_logger import logger
from ..utility.exceptions import InputTypeError

class ModelPreprocess: 

    def generate_time_series(self,
                df: pd.DataFrame,
                date_col: str,
                target_col: str,
                time_granularity: str = "D",
                missing_data_strategy: Union[str,int,dict] = ""
                ) -> pd.DataFrame:
        """
        Sequential execution of transformations to obtain a DataFrame with a time series structure

        Parameters
        ----------
        df: pd.DataFrame
            raw dataframe from which generates timeseries
        date_col: str
            column name identifying the columns to index on
        target_col: str
            column name identifying column from which to generate target
        missing_data_strategy: str or int
            Identifies whether to impute missing values and if so using which strategy/value
            Allowed parameters: 
                if str:
                    "mean" : missing values are replaced with the mean of the known values in the dataset.
                    "median": missing values are replaced with the median of the known values in the dataset.
                    "zero": missing values are replaced with the 0.
                    "bfill": missing values are replaced with the next available value in the dataset.
                    "ffill": missing values are replaced with the most recent preceding value in the dataset.
                if int:
                    replace NaN with the specified integer
                if dict:
                    replace NaN using the specified interpolation method (allowed "polynomial" or "spline") and its order.
        time_granularity: str 
           specifies temporal granularity

        Returns:
            pd.DataFrame: DataFrame identifying the timeseries generated according to specified hierarchies and time span
        """

         # Verify that df is of type pd.DataFrame
        if not isinstance(df, pd.DataFrame):
            logger.info("Input 'df' must be of type pd.DataFrame", important=True)
            raise InputTypeError('datapreparation_utils.generate_time_series')
        # Verify that datetime_index is a non-empty string and a column in df
        if  date_col not in df.columns:
            logger.info("Invalid value for 'date_col'. It must be a column in 'df'", important=True)
            raise InputTypeError('datapreparation_utils.generate_time_series')
        # Verify that target is a non-empty string and a column in df
        if target_col not in df.columns:
            logger.info("Invalid value for 'target'. It must be a column in 'df'", important=True)
            raise InputTypeError('datapreparation_utils.generate_time_series')
        if missing_data_strategy != "":
            if isinstance(missing_data_strategy, str):
                allowed_strategies = ["mean", "median", "zero", "bfill", "ffill", "interpolate"]
                if missing_data_strategy not in allowed_strategies:
                    logger.info("Invalid value for 'missing_data_strategy'. Allowed values: {}".format(allowed_strategies), important=True)
                    raise ValueError("Invalid value for 'missing_data_strategy' in 'datapreparation_utils.generate_time_series'. Allowed values: {}".format(allowed_strategies))
            elif isinstance(missing_data_strategy, int):
                pass
            elif isinstance(missing_data_strategy, dict):
                allowed_interpolation_method = ['polynomial', "spline"]
                if missing_data_strategy['interpolation'] not in allowed_interpolation_method:
                    logger.info("Invalid value for 'missing_data_strategy - interpolation method'. Allowed values: {}".format(allowed_interpolation_method), important=True)
                    raise ValueError("Invalid value for 'missing_data_strategy' in 'datapreparation_utils.generate_time_series'. Allowed values: {}".format(allowed_interpolation_method))
                if not isinstance(missing_data_strategy['order'], int):
                    logger.info("Invalid value for 'missing_data_strategy - order'. It must be int", important=True)
                    raise ValueError("Invalid value for 'missing_data_strategy - order' in 'datapreparation_utils.generate_time_series'. It must be int")
            else:
                logger.info("Invalid type for 'missing_data_strategy'. It must be str or int", important=True)
                raise TypeError("Invalid type for 'missing_data_strategy' in 'datapreparation_utils.generate_time_series'. It must be str or int or dict")
        
        # Raw dataset from which to generate the time serie
        self.df = df.copy()

        logger.info('Generating the timeseries target')

        df_ts = self.df.set_index(date_col).resample(time_granularity.upper()).size().reset_index()
        df_ts.columns = ["Timestamp", "Target"]
        df_ts.loc[df_ts.Target==0, "Target"] = np.nan

        if missing_data_strategy != "":
            if isinstance(missing_data_strategy,dict):
                if missing_data_strategy['interpolation']=="spline" and (df_ts['Target'].notna().sum() < missing_data_strategy['order'] or missing_data_strategy['order'] >5):
                    logger.error('The number of data points must be larger than the spline degree k or k should be 1<=k<=5.')
                    raise ValueError("Invalid value for spline order in 'datapreparation_utils.generate_time_series'")

            fillna = ReplacerNA(missing_data_strategy)
            df_ts = fillna.fit_transform(df_ts)
            df_ts['Target'] = df_ts['Target'].clip(lower=0)

        return df_ts

def code_to_name(cod: pd.Series, convers_dict: Dict) -> pd.Series:

    """
    The function generates a new column converting the code number into a meaningful string

    Parameters
    ----------
    cod: pd.Series
       The code number column
    convers_dict: dict
        The mapping dictionary from code to string

    Returns
    -------
        pd.Series
        Returns the modified column based on the mapping dictionary
    """
    if not isinstance(cod, pd.Series):
        logger.info("Input 'cod' must be of type pd.Series")
        raise InputTypeError('datapreparation_utils.cod_to_name')
    if not isinstance(convers_dict, dict):
        logger.info("Input 'convers_dict' must be of type dict")
        raise InputTypeError('datapreparation_utils.cod_to_name')
                  
    return cod.apply(lambda i: convers_dict[int(i)])


class Normalizer(TransformerMixin, BaseEstimator):

    """Normalization class for time series (between 0 and 1)
    """

    def __init__(self):
        pass

    def fit(self, X: pd.DataFrame) -> Self:

        """Compute value min and max useful to normalize data

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            fitted normalizer
        """

        self.min = X.iloc[:,1].min()
        self.max = X.iloc[:,1].max()

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform normalization between 0 and 1

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            transformed time series
        """

        normalized = (X.iloc[:,1] - self.min) / (self.max - self.min)
        ris = pd.concat([X.iloc[:,0],normalized],axis=1)
        ris.columns = X.columns
        
        return ris
    
class ReplacerNA(TransformerMixin, BaseEstimator):

    def __init__(self, method: Union[str,int, dict] = "") -> Self:

        """class for handling of NA

        Parameters
        ----------
        method : str | int | dict 
            if str specify the method to replace NA value (mean,median,zero), if int specify the value to replace NA value
            if dict specify which interpolation method to use between polynomial and spline and its order
        Returns
        -------
        self : object
        """
        
        self.method = method

    def fit(self, X: pd.DataFrame) -> Self:

        """Compute value useful for replacing NA

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            fitted replacer
        """
        
        if self.method == "mean":
            self.value = X.iloc[:,1].mean()
            self.method_for_df = None
        elif self.method == "median":
            self.value = X.iloc[:,1].median()
            self.method_for_df = None
        elif self.method == "zero":
            self.value = 0
            self.method_for_df = None
        elif self.method == "bfill":
            self.value = None
            self.method_for_df = "bfill"
        elif self.method == "ffill":
            self.value = None
            self.method_for_df = "ffill"
        elif self.method == "interpolate":
            self.value = None
            self.method_for_df = "interpolate"
        elif isinstance(self.method, dict):
            self.value = self.method["order"]
            self.method_for_df = self.method["interpolation"].lower()

        else:
            self.value = self.method
            self.method_for_df = None

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform replacement of missing values

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            transformed time series
        """
        if self.method_for_df in ["polynomial", "spline"]:
            # Create a temporary DataFrame with a DatetimeIndex
            temp_df = pd.DataFrame({X.columns[1]: X.iloc[:, 1]})
            temp_df.index = pd.to_datetime(X.iloc[:, 0])

            # Perform time-based interpolation in the temporary DataFrame
            temp_df.iloc[:, 0] = temp_df.iloc[:, 0].interpolate(method=self.method_for_df, order = self.value)

            # Assign the interpolated values to the original column in the X DataFrame
            X.iloc[:, 1] = temp_df.iloc[:, 0].values
        else:
            X.fillna(self.value, method=self.method_for_df, inplace=True)
        return X
    
class Detrender(TransformerMixin, BaseEstimator):

    def __init__(self, period: int) -> Self:

        """Detrending time series

        Parameters
        ----------
        period : int
            specify period considered for compute additive decomposition

        Returns
        -------
        self : object
        """

        self.period = period


    def fit(self, X: pd.DataFrame) -> Self:

        """Compute additive decomposition useful to detrend time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            fitted detrender
        """

        additive_decomp = seasonal_decompose(X.iloc[:,1], model="additive", period=self.period, extrapolate_trend="freq")
        self.trend = additive_decomp.trend

        return self

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:

        """Perform detrending of time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            transformed time series
        """

        detrend_time_series = X.iloc[:,1] - self.trend
        ris = pd.concat([X.iloc[:,0],detrend_time_series],axis=1)
        ris.columns = X.columns

        return  ris
    
class Deseasoner(TransformerMixin, BaseEstimator):

    def __init__(self, period: int) -> Self:

        """Deseasonalises time series

        Parameters
        ----------
        period : int
            specify period considered for compute additive decomposition

        Returns
        -------
        self : object
        """

        self.period = period


    def fit(self, X: pd.DataFrame) -> Self:

        """Compute additive decomposition useful to deseasonalises time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            fitted deseasoner
        """
        
        additive_decomp = seasonal_decompose(X.iloc[:,1], model="additive", period=self.period, extrapolate_trend="freq")
        self.seasonal = additive_decomp.seasonal

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform deseasonalises of time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            transformed time series
        """

        deseason_time_series = X.iloc[:,1] - self.seasonal
        ris = pd.concat([X.iloc[:,0],deseason_time_series],axis=1)
        ris.columns = X.columns

        return ris

class Differencer(TransformerMixin, BaseEstimator):

    def __init__(self, lag: int) -> Self:

        """Differencing time series
        
        Parameters
        ----------
        lag : int
            differencing time series lag

        Returns
        -------
        self : object
        """

        self.lag = lag

    def fit(self, X: pd.DataFrame) -> Self:

        """Compute value useful to compute differencing time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            fitted normalizer
        """

        self.shape = X.shape[0]
        self.lag_time_series = X.iloc[:self.shape-self.lag,1]
        self.timestamp = X.iloc[self.lag:,0].reset_index(drop=True)

        return self
    
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform differencing time series

        Parameters
        ----------
        X : pd.DataFrame
            dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            transformed time series
        """

        time_series_lagged = X.iloc[self.lag:,1].reset_index(drop=True) - self.lag_time_series
        ris = pd.concat([self.timestamp,time_series_lagged], axis=1)
        ris.columns = X.columns
        
        return ris
     
class OutlierRemover(BaseEstimator, TransformerMixin):
    def __init__(self, lower_threshold_percentile=5, upper_threshold_percentile=95):
        """
        Remove outliers from a dataset by capping values above and below thresholds.

        Parameters:
        -----------
        upper_threshold_percentile : int or float, optional (default=95)
            Percentile threshold above which values will be capped.
        lower_threshold_percentile : int or float, optional (default=5)
            Percentile threshold below which values will be capped.
        """
        self.upper_threshold_percentile = upper_threshold_percentile
        self.lower_threshold_percentile = lower_threshold_percentile
        self.upper_threshold = None
        self.lower_threshold = None

    def fit(self, X, y=None):
        """
        Compute the upper and lower thresholds based on percentiles of the input data.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted OutlierRemover object.
        """
        self.upper_threshold = np.percentile(X, self.upper_threshold_percentile)
        self.lower_threshold = np.percentile(X, self.lower_threshold_percentile)
        return self

    def transform(self, X):
        """
        Cap the values above the upper threshold and below the lower threshold.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_transformed : array-like
            Transformed data with capped outlier values.
        """
        X[X > self.upper_threshold] = self.upper_threshold
        X[X < self.lower_threshold] = self.lower_threshold
        return X
    

class Smoother(BaseEstimator, TransformerMixin):
    def __init__(self, window_size):
        """
        Smoothes a time series by applying a moving average window.

        Parameters:
        -----------
        window_size : int
            Size of the moving average window.
        """
        self.window_size = window_size

    def fit(self, X, y=None):
        """
        Fit the Smoother to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Smoother object.
        """
        return self

    def transform(self, X):
        """
        Smooth the input data by applying a moving average window.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_smoothed : array-like
            Smoothed data obtained by applying the moving average window.
        """
        
        X_smoothed = X.rolling(window=self.window_size, min_periods=1).mean()
        return X_smoothed


class Differentiator(BaseEstimator, TransformerMixin):
    def __init__(self, order):
        """
        Differentiates "order" times a time series by taking differences between consecutive values.

        Parameters:
        -----------
        order : int
            Order of differentiation.
        """
        self.order = order

    def fit(self, X, y=None):
        
        """
        Fit the Differentiator to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Differentiator object.
        """
        return self

    def transform(self, X):
        """
        Apply differentiation to the input data by taking differences between consecutive values.
 

        Parameters:
        -----------
        X : array-like
            Input data.


        Returns:
        --------
        X_transformed : array-like
            Transformed data obtained by taking differences between consecutive values.
        """
        X_transformed = X.copy()
        for _ in range(self.order):
            X_transformed = X_transformed.diff()
        X_transformed = np.nan_to_num(X_transformed, nan=0.0)
        return X_transformed


class LogTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        """
        Logarithmic transformation of data.

        Parameters:
        -----------
        None
        """
        pass

    def fit(self, X, y=None):
        """
        Fit the LogTransformer to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted LogTransformer object.
        """
        return self

    def transform(self, X):
        """
        Apply the logarithmic transformation to the input data.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_transformed : array-like
            Transformed data obtained by applying the logarithmic transformation.
        """
        X_transformed = np.log1p(X)
        return X_transformed

    def inverse_transform(self, X_transformed):
        """
        Reconstruct the original data from the transformed data by applying the inverse logarithmic transformation.

        Parameters:
        -----------
        X_transformed : array-like
            Transformed data.

        Returns:
        --------
        X_reconstructed : array-like
            Reconstructed data obtained by applying the inverse logarithmic transformation.
        """
        X_reconstructed = np.expm1(X_transformed)
        return X_reconstructed


class Normalizer(BaseEstimator, TransformerMixin):
    def __init__(self):
        """
        Normalize data to the range [0, 1].

        Parameters:
        -----------
        None
        """
        pass

    def fit(self, X, y=None):
        """
        Fit the Normalizer to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Normalizer object.
        """
        
        self.min = np.min(X)
        self.max = np.max(X)
        return self

    def transform(self, X):
        """
        Normalize the input data to the range [0, 1].

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_normalized : array-like
            Normalized data obtained by scaling the values to the range [0, 1].
        """
        X_normalized = (X - self.min) / (self.max - self.min)
        return X_normalized

    def inverse_transform(self, X_normalized):
        """
        Reconstruct the original data from the normalized data by applying the inverse transformation.

        Parameters:
        -----------
        X_normalized : array-like
            Normalized data.

        Returns:
        --------
        X : array-like
            Reconstructed data obtained by applying the inverse transformation.
        """
        X = X_normalized * (self.max -self.min) + self.min
        return X



import pytest
import pandas as pd
import geopandas as gpd
import os

from tc_uc4.datahandler.datahandler import DataHandler
from tc_uc4.utility.exceptions import InputTypeError

@pytest.fixture
def data_handler(tmp_path):
    return DataHandler(data_folder=str(tmp_path))

def test_data_handler_configured_init_read(tmp_path):
    # Create a Pandas DataFrame
    data = {'Column1': [1, 2, 3], 'Column2': ['A', 'B', 'C']}
    df = pd.DataFrame(data)

    # Write DataFrame to a Parquet file
    file_path = tmp_path / 'test.parquet'
    df.to_parquet(file_path)

    DH = DataHandler(data_folder=str(tmp_path))
    df_check = DH.read(filename='test.parquet', folder='')
    df_check2 = DH.read(filename='test.parquet', folder='.')

    assert type(df_check) == pd.DataFrame
    assert type(df_check2) == pd.DataFrame
    assert df.equals(df_check)
    assert df.equals(df_check2)

    # Can't find the file
    with pytest.raises(SystemExit):
        DH.read(filename='test.parquet', folder='/')
    
    # Can't find the file
    with pytest.raises(SystemExit):
        DH.read(filename='test.parquet', folder='-')


def test_datahandler_wrong_init():
    with pytest.raises(InputTypeError):
        DH = DataHandler(data_folder=123)

def test_datahandler_read_wrong_input_types(data_handler):
    with pytest.raises(SystemExit):
        data_handler.read(123,'string ok',True)

def test_datahandler_read_geodataframe_wrong_input_types(data_handler):
    with pytest.raises(SystemExit):
        data_handler.read_geodataframe('string ok', 123, True)

def test_datahandler_write_wrong_input_types(data_handler):
    with pytest.raises(SystemExit):
        data_handler.write(pd.Series, 'string ok', 'not a bool')

def test_datahandler_read_not_supp_extension(data_handler):
    df = pd.DataFrame([1])
    with pytest.raises(SystemExit):
        data_handler.read('this_is_a_file.txt')

def test_datahandler_read_geodataframe_not_supp_extension(data_handler):
    df = pd.DataFrame([1])
    with pytest.raises(SystemExit):
        data_handler.read_geodataframe('this_is_a_file.txt')

def test_datahandler_write_not_supp_extension(data_handler):
    df = pd.DataFrame([1])
    with pytest.raises(SystemExit):
        data_handler.write(df, 'this_is_a_file.txt')

def test_datahandler_write_folder(data_handler):
    df = pd.DataFrame([1])
    with pytest.raises(SystemExit):
        data_handler.write(df, 'test.parquet', 'pippo')

def test_datahandler_outputs(tmp_path, data_handler):
    data = {'Column1': [1, 2, 3], 'Column2': ['A', 'B', 'C']}
    df = pd.DataFrame(data)

    data_geo = {
    'City': ['New York', 'Los Angeles', 'Chicago'],
    'Latitude': [40.7128, 34.0522, 41.8781],
    'Longitude': [-74.0060, -118.2437, -87.6298]
    }
    # Create GeoDataFrame
    geo_df = gpd.GeoDataFrame(data_geo, geometry=gpd.points_from_xy(data_geo['Longitude'], data_geo['Latitude']))
    
    os.mkdir(os.path.join(tmp_path, 'regions'))
    file_path_geo = tmp_path / 'regions/test.parquet'
    geo_df.to_parquet(file_path_geo)

    file_path_pd = tmp_path / 'test.parquet'
    df.to_parquet(file_path_pd)

    df = data_handler.read('test.parquet', folder='')
    assert type(df) == pd.DataFrame

    df = data_handler.read_geodataframe('test.parquet', folder='regions')
    assert type(df) == gpd.GeoDataFrame



from __future__ import annotations

import itertools

import numpy as np
import pandas as pd
import geopandas as gpd

from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.spatial.distance import squareform

import matplotlib.pyplot as plt
from dtaidistance import dtw
from dtaidistance import dtw_visualisation as dtwvis

import matplotlib.pyplot as plt
import plotly
import plotly.express as px
import plotly.graph_objects as go

from typing_extensions import Self

from ..utility.resources import log_exception
from ..utility.tele_logger import logger

from ..utility.constants import code_to_region_name


class DataVisualization:

    def __init__(
        self,
        use_case: str,
        df: pd.DataFrame,
        target: str = "volumes"
    ) -> Self:
        
        """Initializing the DataVisualization class.
            A class for the visualization of general situation descript by dataframe (the class consider all value,
            for each aggregation level).

        Parameters
        ----------
        use_case : str
            specify the name of use case (necessary for the plot title)
        df : pd.DataFrame
            the original dataset
        target : str
            the name of column target, by default "volumes"
        dict_reg : dict
            a dictionary that convert codice_regione to name , by default None
        """

        self.use_case = use_case
        self.dict_regioni = code_to_region_name
        self.df = df
        self.target = target

        self.df[target] = 1

    @log_exception(logger)
    def conteggi(
        self,
        level1: str,
        level2: str = None,
        level3: str = None,
        freq: str = None,
    ) -> pd.Series:
        
        """Compute the aggregation by specify level

        Parameters
        ----------
        target : str
            name of target column
        level1 : str
            name of first level aggregation
        level2 : str
            name of second level aggregation, by default None
        level3: str
            name of third level aggregation, by default None

        Returns
        -------
        pd.Series
            Series of aggregation count

        Raises
        ------
        Exception
            _description_
        """

        # try:
        #     pd.Period(data_inizio)
        #     pd.Period(data_fine)
        # except:
        #     raise Exception("La data non è valida")

        livelli = list(filter(None, [level1, level2, level3]))
        livelli1 = [
            pd.Grouper(key=i, freq=freq) if self.df[i].dtype == "<M8[ns]" else i
            for i in livelli
        ]
        regioni = list(self.dict_regioni.values())

        logger.info(msg="START elaborate counting dataframe", important=True)

        # select DATA
        # i = " 23:59:59"
        # tab = self.df.query(f"({data} >= '{data_inizio}') & ({data} <= '{data_fine + i}')")

        # aggregazione per livelli specificati
        ris = (
            self.df.groupby(livelli1, as_index=False)[self.target]
            .sum()
            .sort_values([self.target], ascending=0)
            .set_index(livelli)
        )

        # regioni mancanti
        if self.df[level1].unique()[0] in regioni:
            regioniMancanti = np.setdiff1d(
                regioni, ris.index.get_level_values(0).unique()
            )
            if len(livelli) > 1:
                regioniMancanti = [(i, "null") for i in regioniMancanti]
        else:
            regioniMancanti = []

        zero = pd.Series(np.nan, index=regioniMancanti)

        if ris.shape[0] > 1:
            ris = pd.concat([ris.squeeze(), zero])
        else:
            ris = pd.concat([ris, zero])[self.target]

        ris = ris.to_frame()

        # indici e colonne
        if len(livelli) > 1:
            ris.index = pd.MultiIndex.from_tuples(ris.index, names=(i for i in livelli))

        ris.reset_index(inplace=True)
        ris.columns = livelli + [self.target]

        #self.cont = ris

        logger.info(msg="DONE elaborate counting dataframe", important=True)

        return ris

    def plot(
        self,
        df: pd.DataFrame,
        level1: str,
        level2: str = None,
        level3: str = None,
    ) -> plotly.graph_objs._figure.Figure:
        
        """Sunburst plot of aggregate data

        Parameters
        ----------
        df : pd.DataFrame
            dataframe of aggregate data (result of conteggi function)
        target : str
            name of target column (count column)
        level1 : str
            name of first level aggregation
        level2 : str, optional
            name of second level aggregation, by default None
        level3 : str, optional
            name of third level aggregation, by default None

        Returns
        -------
        plotly.graph_objs._figure.Figure
            sunburst plot
        """

        livelli = [level1, level2, level3]
        levels = list(filter(None, livelli))
        name_levels = " e ".join(levels).replace("_", " ").title()

        fig = px.sunburst(
            df,
            path=levels,
            values=self.target,
            width=600,
            height=600,
            title=f"Numero di {self.use_case} per {name_levels}",
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )
        fig.update_traces(textinfo="label+percent parent")

        return fig

    def boxplot(self, df: pd.DataFrame, variable: str):
        
        """Boxplot of aggregated data

        Parameters
        ----------
        df : pd.DataFrame
            dataframe of aggregate data (result of conteggi function)
        variable : str
            name of variable on x axis 

        Results
        -------
        plotly.graph_objs._figure.Figure
            boxplot
        """

        fig = px.box(
            df,
            x=variable,
            y=self.target,
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )

        return fig

    def DTW(self, df: pd.DataFrame, column: str, data: str, val1: str, val2: str) -> tuple:

        """compute distance between two time series

        Parameters
        ----------
        df : pd.DataFrame
            dataframe result of conteggi function
        column : str
            name of one column aggregate on
        data : str
            name of data column
        val1: str
            name of column modality
        val2 : str
            name of column modality

        Returns
        -------
        tuple
            tuple containing a figure that representing distance between time series and the value of distance

        """

        df_sort = df.sort_values(data)
        x = df_sort[df_sort[column] == val1][self.target].values
        y = df_sort[df_sort[column] == val2][self.target].values       

        # Normalize time series
        x = x / x.max()
        y = y / y.max()

        distance = dtw.distance_fast(x, y, use_pruning=True)
        path = dtw.warping_path(x, y)

        fig, ax = plt.subplots(2, 1, figsize=(20, 9))
        dtwvis.plot_warping(x, y, path, fig=fig, axs=ax)

        ax[0].set_title(
            f"Dynamic Time Warping, {column} ({val1}-{val2})",
            fontsize=20,
            y=1.1,
            color="gray",
            loc="left",
            family="serif",
        )
        ax[1].set_xlabel("time", fontsize=12)
        ax[0].set_ylabel(self.target, fontsize=12)
        ax[1].set_ylabel(self.target, fontsize=12)
        fig.tight_layout()

        normalized_distance = distance / np.sqrt(len(x) * len(y))

        return (fig, normalized_distance)
    
    def DTW_distance_matrix(self, df: pd.DataFrame, column: str, data: str) -> tuple:

        """computes DTW distance matrix of all sub time series of a column

        Parameters
        ----------
        df : pd.DataFrame
            dataframe result of conteggi function
        column : str
            name of one column aggregate on
        data : str
            name of data column

        Returns
        -------
        tuple
            tuple containing the matrix and the plot of matrix
        """

        df_sort = df.sort_values(data)

        lista = [(df_sort[df_sort[column] == i][self.target].values.astype(np.double),i)
         for i in df_sort[column].unique()]

        lista_time_series = [i[0]/i[0].max() for i in lista]
        len_time_series = np.array([len(i[0]) for i in lista])
        len_matrix_time_series = np.sqrt(np.einsum("i, j -> ij", len_time_series, len_time_series))
        distance_matrix = dtw.distance_matrix_fast(lista_time_series)
        
        distance_matrix_normalized = distance_matrix / len_matrix_time_series

        fig = px.imshow(distance_matrix_normalized, text_auto=True, template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="grey"),
                    font=dict(color="green"),
                width=800,
                height=700,
                title=f"Distance matrix ({column})",
                )))

        fig.update_layout(
            xaxis = dict(
                tickmode = 'array',
                tickvals = [i for i in range(len(lista))],
                ticktext = [i[1] for i in lista],tickfont=dict(family='Rockwell', color='green', size=10)
            ),
            yaxis = dict(
                tickmode = 'array',
                tickvals = [i for i in range(len(lista))],
                ticktext = [i[1] for i in lista],tickangle=-45,tickfont=dict(family='Rockwell', color='green', size=10)
            )           
        )

        return (fig,distance_matrix_normalized,lista)

    def plot_distance_matrix(self, distance_matrix: np.ndarray, lista: list[tuple]):
        
        dists = squareform(distance_matrix)
        linkage_matrix = linkage(dists, "ward",optimal_ordering=0)
        
        dendr = plt.figure(figsize=(10,6))
        dendrogram(linkage_matrix, labels=[i[1] for i in lista],leaf_rotation=0,orientation="right")
        plt.title("DTW distance - dendrogramm")
        plt.tick_params(
            axis='x',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            labelbottom=False) # labels along the bottom edge are off
        for pos in ['right', 'top', 'bottom', 'left']:
            plt.gca().spines[pos].set_visible(False)
        
        return dendr

    def add_geographical_info(
        self, target: str, region: str, df_geo: gpd.GeoDataFrame, region_geo: str
    ) -> gpd.GeoDataFrame:
        
        """Aggregates df on "spatial_variable" and joins with the "regions" geodataframe on "join_variable"

        Parameters
        ----------
        target : str
            name of target column
        region : str
            name of region column
        df_geo : gpd.GeoDataFrame
            GeoDataFrame containing geometry for Italian region
        region_geo : str
            name of region column
        data : str
            name of data column
        data_inizio : str
            initial data to select
        data_fine : str
            final data to select

        Returns
        -------
        gpd.GeoDataFrame
            geodataframe containing geographical information and the aggregated "spatial_variable"
        """

        df_aggreg = self.conteggi(region)
        df_aggreg.set_index(region, inplace=True)
        df_aggreg = df_aggreg.reindex(df_geo[region_geo])

        df_geo[target] = df_aggreg.values

        return df_geo

    def plot_geo(
        self,
        df: gpd.GeoDataFrame,
        target: str,
        region_geo: str,
        confronto: str = None,
        express: str = None,
        title: str = "",
    ) -> plotly.graph_objs._figure.Figure:
        
        """Territorial plot of aggregated data for region

        Parameters
        ----------
        df : gpd.GeoDataFrame
            dataframe with gegographical info join with result of conteggi function
        target : str
            name of target column
        region_geo : str
            name of column with geographical info
        confronto : str 
            name of column with comparision value with target column (for example number of population),by default None
        express : str 
            expression that describe relationship between confronto column and target column
        title : str 
            title for the plot

        Results
        -------
        plotly.graph_objs._figure.Figure
            geographical plot
        """

        def generateColorScale(colors, naColor):
            colorArray = []
            colorArray.append([0, naColor])
            for grenze, color in zip(np.linspace(0.03, 1, len(colors)), colors):
                colorArray.append([grenze, color])
            return colorArray

        df = df.set_index(region_geo)
        df = df.fillna(0)

        target_name = target
        confronto_name = confronto

        if confronto:
            target = df[target_name]
            confronto = df[confronto_name]
            df[target_name] = eval(express).values

        fig = px.choropleth(
            df,
            geojson=df["geometry"],
            locations=df.index,
            color=target_name,
            projection="mercator",
            title=title,
            width=700,
            height=700,
            color_continuous_scale=generateColorScale(
                colors=["blue", "red"], naColor="grey"
            ),
        )
        fig.update_geos(fitbounds="locations", visible=False)

        return fig



import os
import subprocess

import mlflow
import pytest

from tc_uc4.datahandler.datahandler import DataHandler
from tc_uc4.utility.constants import predict_dataframe_columns
from tc_uc4.utility.resources import get_configuration

models = ["Abruzzo", "Abruzzo_F", "Abruzzo_M", "Basilicata", "Basilicata_F", "Basilicata_M", "Bolzano",
          "Bolzano_F", "Bolzano_M", "Calabria", "Calabria_F", "Calabria_M", "Campania", "Campania_F",
          "Campania_M", "Emilia-Romagna", "Emilia-Romagna_F", "Emilia-Romagna_M", "Friuli-Venezia-Giulia",
          "Friuli-Venezia-Giulia_F", "Friuli-Venezia-Giulia_M", "Italia", "Lazio", "Lazio_F", "Lazio_M",
          "Liguria", "Liguria_F", "Liguria_M", "Lombardia", "Lombardia_F", "Lombardia_M", "Molise",
          "Molise_F", "Molise_M", "Piemonte", "Piemonte_F", "Piemonte_M", "Puglia", "Puglia_F",
          "Puglia_M", "Sardegna", "Sardegna_F", "Sardegna_M", "Sicilia", "Sicilia_F", "Sicilia_M",
          "Toscana", "Toscana_F", "Toscana_M", "Trento", "Trento_F", "Trento_M", "Umbria", "Umbria_F",
          "Umbria_M", "Valle D Aosta", "Valle D Aosta_F", "Valle D Aosta_M", "Veneto", "Veneto_F", "Veneto_M"
          ]

expected_metrics = {
    "MAE": 1.557,
    "MAPE": 0.489,
    "MSE": 6.01,
    "Percentage Coverage": 80.66,
    "R2": -0.134,
    "RMSE": 2.098,
}


def test_job_experiment():
    experiment_name = "unittest"
    python_cmd = 'python'
    
    project_path = os.path.split(os.path.dirname(os.path.abspath(__file__)))[0]
    config_path = os.path.join(project_path, "tests/config")
    path_configs = experiment_name + "_path.toml"
    model_configs = experiment_name + "_prophet.toml"

    print(os.path.join(project_path, "experiments", "train_and_tune_prophet.py"))

    process = subprocess.Popen(f'{python_cmd} {os.path.join(project_path, "experiments", "train_and_tune_prophet.py")} -E {experiment_name}', shell=True)

    process.wait()

    experiment_folder = get_configuration("data_paths", config_path, path_configs)["experiment_folder"]
    experiment_folder = os.path.join(project_path, experiment_folder)
    model_name = get_configuration("model", config_path, model_configs)["model_name"]

    mlflow.set_tracking_uri(experiment_folder)
    experiment_id = mlflow.get_experiment_by_name(experiment_name).experiment_id

    # Find best run
    run = mlflow.search_runs(experiment_ids=experiment_id).loc[0]

    # Test if trained models are present in artifacts
    model_uri = f"runs:/{run['run_id']}/{model_name}.pkl"
    model = mlflow.sklearn.load_model(model_uri)

    # Test model in models
    assert len(model.models_dict) == len(models)
    for id_pred, m in model.models_dict.items():
        assert id_pred.replace("/", "_") in models

        # Test parameters in specified range
        if m is not None:
            assert m.model.weekly_seasonality in [True, False]
            assert m.model.yearly_seasonality in [True, False]

    # Test model metrics
    for metric, expected_value in expected_metrics.items():
        assert pytest.approx(run[f"metrics.{metric}"], 0.1) == expected_value

    ################################################################
    # Test Job
    ################################################################

    predict_save_folder = get_configuration("data_paths", config_path, path_configs)["predict_save_folder"]
    process = subprocess.Popen(f'{python_cmd} {os.path.join(project_path, "jobs", "predict_prophet.py")} -E {experiment_name}',shell=True)
    process.wait()

    DH = DataHandler(predict_save_folder)
    predictions = DH.read("prophet_predictions.parquet", folder="")

    # Test predictions df shape
    assert len(predictions.columns) == len(predict_dataframe_columns)
    assert len(predictions) == 6100
    for col in predictions.columns:
        assert col in predict_dataframe_columns

    for row in predictions.iterrows():
        if row[1] is None:
            assert row[2] == 0.0
            assert row[3] == 0.0
            assert row[4] == 0.0
            assert row[5] == 0.0

    # Delete created folders
    process = subprocess.Popen(
        f'rm -rf {os.path.join(project_path, "tests/experiments")}',
        shell=True,
    )
    process = subprocess.Popen(
        f'rm -rf {os.path.join(project_path, "tests/data/output")}',
        shell=True,
    )



import warnings;
warnings.simplefilter('ignore')
import pandas as pd
import numpy as np
from sklearn.metrics import mean_squared_error, r2_score, mean_absolute_error, mean_absolute_percentage_error

from ..utility.exceptions import ColumnNotFound
from ..utility.tele_logger import logger


def forecast_coverage(df_real : pd.DataFrame, df_pred : pd.DataFrame) -> float:
    """
    Calculate the forecast coverage metric.

    Parameters:
    ----------
    df_real: pd.DataFrame
        DataFrame of real values with columns 'Timestamp' and 'Target'.
    df_pred: pd.DataFrame
        DataFrame of predicted values with columns 'Timestamp', 'Pred_mean',
        'Pi_lower_95', and 'Pi_upper_95'.

    Returns:
    -------
    Float
        Forecast coverage as a percentage.
    """
    # Check if the actual values fall within the prediction intervals
    within_interval = (df_real['Target'] >= df_pred['Pi_lower_95']) & (df_real['Target'] <= df_pred['Pi_upper_95'])
    
    # Calculate the forecast coverage as the percentage of actual values within the prediction intervals
    forecast_coverage = within_interval.mean() * 100
    
    return forecast_coverage


def evaluations(df_real: pd.DataFrame, df_pred: pd.DataFrame, date: str = 'Timestamp' , y_true: str = 'Target', y_pred: str = 'Pred_mean') -> pd.DataFrame:
    """
    Create a DataFrame representing the evaluation metrics

    Parameters
    ----------
    df_real: pd.DataFrame
        the dataframes containing the real values
    df_pred: pd.DataFrame
        the dataframes containing the predicted values
    date: str
        the column name for the timestamp
    y_true: str
        the target column name for the real datasets
    y_pred: str
        the target column name for the predict datasets

    Returns
    -------
        A pd.DataFrame containing the results for MAE, RMSE, MSE and R2
    """
    if date not in df_real.columns:
        raise ColumnNotFound(func='evaluations', column=date, data=df_real)
    if date not in df_pred.columns:
        raise ColumnNotFound(func='evaluations', column=date, data=df_pred)
    if y_true not in df_real.columns:
        raise ColumnNotFound(func='evaluations', column=y_true, data=df_real)
    if y_pred not in df_pred.columns:
        raise ColumnNotFound(func='evaluations', column=y_pred, data=df_pred)

    logger.info('START Ealuating results')

    df_real= df_real.set_index(date).dropna()
    df_pred= df_pred.set_index(date).dropna()
    inter = df_real.index.intersection(df_pred.index)

    if len(inter) == 0:
        logger.warning(f'Length of intersection between Predictions and Ground Truth Datasets for {df_pred["Id_pred"][0]} is Null. Probably one of the two Datasets has not available data (all NaN).')
        scores = {'Id_pred': df_pred["Id_pred"][0], 'MAE':[np.nan], 'MAPE':[np.nan], 'RMSE': [np.nan], 'MSE':[np.nan], 'R2':[np.nan], 'Percentage Coverage':[np.nan]}
    else:    
        df_real = df_real.loc[inter]
        df_pred = df_pred.loc[inter]

        # Mean Absolute Error
        logger.info('Evaluating MAE')
        mae = round(mean_absolute_error(df_real[y_true], df_pred[y_pred]),2)
        # Mean Absolute Error
        logger.info('Evaluating MAPE')
        mape = round(mean_absolute_percentage_error(df_real[y_true], df_pred[y_pred]),2)
        # Root Mean Square Error
        logger.info('Evaluating RMSE')
        rmse = round(mean_squared_error(df_real[y_true], df_pred[y_pred], squared = False),2)
        # Mean Square Error
        logger.info('Evaluating MSE')
        mse = round(mean_squared_error(df_real[y_true], df_pred[y_pred]),2)
        # Coefficient of Determination
        logger.info('Evaluating R2')
        r2 = round(r2_score(df_real[y_true], df_pred[y_pred]),2)
        # Forecast Coverage Percentage
        logger.info('Evaluating Forecast Coverage Percentage')
        coverage = forecast_coverage(df_real=df_real, df_pred=df_pred)

        scores = {'Id_pred': df_pred["Id_pred"][0], 'MAE':[mae], 'MAPE':[mape], 'RMSE': [rmse], 'MSE':[mse], 'R2':[r2], 'Percentage Coverage':[coverage]}

        logger.info('DONE Evaluating results')

    # Saving the scores in a DF
    return pd.DataFrame(scores)



import pandas as pd
import numpy as np
import pytest
from tc_uc4.analytics.evaluationMetrics import evaluations
from tc_uc4.utility.exceptions import ColumnNotFound

def test_evaluations():
     # Sample data 
     real_df = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-26', '2023-06-27', '2023-06-28', '2023-06-29']),
                             'Target': [10, 20, 30, 40]})
     prophet_df = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-26', '2023-06-27', '2023-06-28', '2023-06-29']),
                                'Id_pred' : ['TEST', 'TEST', 'TEST', 'TEST'],
                                'Pred_mean': [12, 18, 32, 37],
                                'Pi_upper_95': [14, 19, 37, 40],
                                'Pi_lower_95': [10, 15, 29, 33]})

     # Calculate evaluation metrics
     result = evaluations(real_df, prophet_df)

     expected_result = pd.DataFrame({'Id_pred': {0: 'TEST'},
                                     'MAE': {0: 2.25},
                                     'MAPE': {0: 0.11},
                                     'RMSE': {0: 2.29},
                                     'MSE': {0: 5.25},
                                     'R2': {0: 0.96},
                                     'Percentage Coverage': {0: 75.0}})

     # Verify the content of the evaluation dataframe
     assert isinstance(result, pd.DataFrame)  # The result should be a DataFrame
     assert result.shape == (1,7)  # The DataFrame should have 4 rows
     assert set(result.columns) == {'Id_pred', 'MAE', 'MAPE', 'MSE', 'R2', 'RMSE', 'Percentage Coverage'}  # The DataFrame should have the correct columns
     pd.testing.assert_frame_equal(result, expected_result)

def test_evaluations_column_not_found():
    # Sample data without the 'Pred_mean' column in the pred dataframe
     df_real = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-27', '2023-06-28', '2023-06-29']),
                             'Target': [10, 20, 30]})
     df_pred = pd.DataFrame({'Timestamp': pd.to_datetime(['2023-06-27', '2023-06-28', '2023-06-29']),
                             'Id_pred' : ['TEST', 'TEST', 'TEST'],
                             'Value': [12, 18, 32]})

     # Sample data without the 'Timestamp' column in the real dataframe
     df_real2 = pd.DataFrame({'timestamp': pd.to_datetime(['2023-06-27', '2023-06-28', '2023-06-29']),
                              'Target': [10, 20, 30]})

     # Check if the ColumnNotFound exception is raised
     with pytest.raises(ColumnNotFound):
          evaluations(df_real, df_pred)
     
     with pytest.raises(ColumnNotFound):
          evaluations(df_real2, df_pred)



from typing import Any, Dict

import pandas as pd
from prophet import Prophet
from typing_extensions import Self

from ..analytics.evaluationMetrics import evaluations
from ..datapreparation.datapreparation_utils import logger
from ..datapreparation.prep import PreprocessingClass
from ..utility.experiment_utils import create_zero_dataframe
from ..utility.constants import predict_dataframe_columns
from .algorithm import Prophet_model, prophet_tuning
from .prophet_utils import preprocess_prophet_input, preprocess_prophet_output
from copy import deepcopy

class ProphetExperiment:
    def __init__(self):
        pass
    
    def fit(self, df_train_dict: Dict, df_val_dict: Dict, 
            hyperparameters_grid: Dict[str, Any], max_na_ratio: float = 0.5,) -> Self:
        
        count = 0
        self.models_dict = {}
        self.models_best_params = {}
        self.max_na_ratio = max_na_ratio

        # Train and tune all models for all the possible levels in the hierarchy
        tmp_df_train_dict = deepcopy(df_train_dict)
        tmp_df_val_dict = deepcopy(df_val_dict)

        for id_pred in tmp_df_train_dict.keys():
            self.models_dict[id_pred] = None
            self.models_best_params[id_pred] = {}

            logger.info(f"START training {id_pred}")
            prophet_train_df = preprocess_prophet_input(tmp_df_train_dict[id_pred], date = "Timestamp", target = "Target")
            prophet_val_df = preprocess_prophet_input(tmp_df_val_dict[id_pred], date = "Timestamp", target = "Target")

            if (prophet_train_df['y'] > 0).sum() / len(prophet_train_df) < self.max_na_ratio:
                logger.info(f"The train set for time series {id_pred} has more than {self.max_na_ratio*100}% null values, skipping it.")
                count += 1
                logger.info(f"Remaining number of iteration - {len(tmp_df_train_dict.keys())-count}")
                continue
            try:
                # Tune if we have a non-empty validation set
                best_params = prophet_tuning(hyperparameters_grid, prophet_train_df, prophet_val_df)
                Model = Prophet_model(best_params)
                # Retrain on train and val data with best parameters
                Model.fit(pd.concat([prophet_train_df, prophet_val_df]))

                self.models_best_params[id_pred] = best_params
                self.models_dict[id_pred] = Model
            except ValueError as e:
                logger.info(f"Skipping training {id_pred}, due to: {e}")
                count += 1
                logger.info(f"Remaining number of iteration - {len(tmp_df_train_dict.keys())-count}")
                continue

            logger.info(f"DONE training {id_pred}")

            count += 1
            logger.info(f"Remaining number of iteration - {len(tmp_df_train_dict.keys())-count}")

        return self
    
    def refit(self, df_dict : Dict):
                
        count = 0
        # Train and tune all models for all the possible levels in the hierarchy
        tmp_df_dict = deepcopy(df_dict)

        for id_pred in tmp_df_dict.keys():
            logger.info(f"START Re-Training {id_pred}")
            prophet_df = preprocess_prophet_input(tmp_df_dict[id_pred], date = "Timestamp", target = "Target")

            # If the models exists and have been trained, execute the retrain on the whole dataset
            if isinstance(self.models_dict[id_pred], Prophet_model):
                Model = Prophet_model(self.models_best_params[id_pred])
                # Retrain on train and val data with best parameters
                Model.fit(prophet_df)

                self.models_dict[id_pred] = Model

            logger.info(f"DONE Re-Training {id_pred}")

            count += 1
            logger.info(f"Remaining number of iteration - {len(tmp_df_dict.keys())-count}")

        return self

    def create_hyperparameters_table(
        self, hyperparameters: dict[str, Any]) -> dict[str, Any]:
        def get_params_to_log(
            model: Prophet, hyperparameters: list[str]) -> dict[str, Any]:
            return {hyper: getattr(model, hyper) for hyper in hyperparameters}

        return {id_pred.replace('/', '_') : get_params_to_log(model.model, list(hyperparameters.keys())) if model else None 
                for id_pred, model in self.models_dict.items()}

    def predict(self, df_test_dict: dict) -> pd.DataFrame:
        predictions = []
        tmp_df_test_dict = deepcopy(df_test_dict)
        for id_pred, df_test in tmp_df_test_dict.items():
            prophet_df_test = preprocess_prophet_input(df_test, date = "Timestamp", target = "Target")
            model = self.models_dict[id_pred]
            if model is None:
                logger.info(f"Skipping prediction for {id_pred}, as the model has not been trained")
                df_output = create_zero_dataframe(predict_dataframe_columns, len(prophet_df_test))
                df_output['Id_pred'] = id_pred
                df_output['Timestamp'] = list(prophet_df_test['ds'])
            else:
                try:
                    df_test_pred = model.predict(prophet_df_test)
                    df_output = preprocess_prophet_output(df_test_pred, id_pred)
                except ValueError as e:
                    logger.info(f"Skipping prediction for {id_pred}, due to: {e}")
                    df_output = create_zero_dataframe(
                        predict_dataframe_columns,
                        len(prophet_df_test),
                    )
                    df_output['Id_pred'] = id_pred
                    df_output['Timestamp'] = list(prophet_df_test['ds'])
            # Evaluate model
            predictions.append(df_output)
            logger.info(f"DONE predicting {id_pred}")

        return pd.concat(predictions)

    def evaluate(self, df_test_dict: dict) -> pd.DataFrame: 
        metrics = []
        tmp_df_test_dict = deepcopy(df_test_dict)
        predictions = self.predict(df_test_dict)
        
        for id_pred, df_test in tmp_df_test_dict.items():
            logger.info(f"START evaluating {id_pred}")           
            df_pred = predictions[predictions["Id_pred"] == id_pred]
            if df_pred.Pred_mean.sum() > 0:
                metrics.append(evaluations(df_test, df_pred, date='Timestamp' , y_true='Target', y_pred='Pred_mean'))  
        
        metrics = pd.concat(metrics).dropna()
        return metrics



from typing import Dict, Tuple
import pandas as pd
import numpy as np
from sklearn.metrics import mean_squared_error, mean_absolute_percentage_error
# Hierarchical Forecast
from hierarchicalforecast.core import HierarchicalReconciliation
from hierarchicalforecast.utils import aggregate

from ..utility.tele_logger import logger
from ..analytics.evaluationMetrics import evaluations



def get_hierarchical_df(df: pd.DataFrame, hierarchy: Dict[str, str],
                        time_granularity: str, date_col: str, target_col: str) -> pd.DataFrame:
    """
    The function generates the dataframe suitable for obtaining 
    HierarchicalForecast info. It generates this dataframe
    starting from the input dataframe by using the information
    of the configuration file.

    Parameters
    ----------
    df: pd.DataFrame
        The input dataframe (typically inside .../data/input)
    preprocessor: PreprocessingClass
        The initialized preprocessing class
    hierarchy: dict
        Dictionary retrieved from the configuration file.
        it contains the hierarchy columns ordered by levels
    time_granularity: str
        string referring to the time granularity.
        Information retrieved from the configuration file.

    Returns
    -------
        pd.DataFrame
        A dataframe of columns: ds|y|level0|level1|level2|level3
        level0 refers to Italia
        level1, level2, level3 are obtained from the hierarchy dict
    """
    
    output_df = df[[date_col, target_col]]
    output_df['level0'] = 'Italia'
    output_df['y'] = 1
    output_df[date_col] = output_df[date_col].dt.to_period(time_granularity).dt.to_timestamp()
    for level, col_name in hierarchy.items():
        output_df[level] = df[col_name]
    output_df = output_df.rename(columns={date_col:'ds'})
    return output_df

def get_hierarchical_info(hier_df: pd.DataFrame) -> Tuple[pd.DataFrame, dict]:
    """
    The function generates the hierarchical aggregation info
    starting from the hierarchical dataframe obtained by <get_hierarchical_df>

    Parameters
    ----------
    hier_df: pd.DataFrame
        Dataframe obtained by <get_hierarchical_df>

    Returns
    -------
        Tuple(S_df: pd.DataFrame, tags: dict)
        S_df binary matrix to understand the hierarchy levels
        tags maps the hierarchy level to the hierarchy values 

    """
    filtered_columns = [col for col in hier_df.columns if 'level' in col]
    filtered_columns.sort()
    spec = [filtered_columns[:i] for i in range(1, len(filtered_columns) + 1)]
    Y_df, S_df, tags = aggregate(df = hier_df, spec = spec)
    return Y_df, S_df, tags

def get_hier_pred(test_pred):
    """
    _summary_

    Parameters
    ----------
    test_pred
        _description_

    Returns
    -------
        _description_
    """
    hier_pred = test_pred.rename(columns = {
        'Timestamp' : 'ds', 'Pred_mean' : 'Prophet'
    })

    hier_pred = hier_pred.set_index('Id_pred')
    hier_pred = hier_pred.rename_axis('unique_id')
    for col in hier_pred.columns:
        if col not in ['ds', 'Prophet']:
            hier_pred = hier_pred.drop(col, axis = 1)
    return hier_pred

def rename_index(df):
    """
    _summary_

    Parameters
    ----------
    df
        _description_

    Returns
    -------
        _description_
    """
    unique_indexes = list(df.index.unique())
    for index in unique_indexes:
        if index != 'Italia':
            new_index = 'Italia/' + index
            df.rename(index = {index : new_index}, inplace = True)
    return df

def concat_dataframes(dict_dataframe):
    """
    _summary_

    Parameters
    ----------
    dict_dataframe
        _description_

    Returns
    -------
        _description_
    """
    dataframes_to_concat = []
    for id_pred, df in dict_dataframe.items():
        df['Id_pred'] = id_pred
        dataframes_to_concat.append(df)
    concat_dataframe = pd.concat(dataframes_to_concat)
    concat_dataframe = concat_dataframe.set_index('Id_pred')
    concat_dataframe = concat_dataframe.rename_axis('unique_id')
    return concat_dataframe

def check_date_consistency(hier_train, concat_df_train):
    return np.all(pd.to_datetime(hier_train['ds']).values == pd.to_datetime(concat_df_train['Timestamp']).values)

def hierarchical_reconcilers(hier_pred, hier_train, sum_df, tags, 
                             reconcilers, confidence_level = 95, method = 'bootstrap'):
    
    hrec = HierarchicalReconciliation(reconcilers=reconcilers)
    Y_rec_df = hrec.reconcile(Y_hat_df = hier_pred, Y_df = hier_train, S = sum_df,
                            tags = tags, level=[confidence_level],
                            intervals_method = method)
    
    numeric_columns = Y_rec_df.select_dtypes(include=[np.number]).columns
    Y_rec_df[numeric_columns] = Y_rec_df[numeric_columns].round().astype(int) # update a round before astype (as astype is like "floor")
    return Y_rec_df, hrec.reconcilers

def best_reconcile(concat_df_test, reconciled_df, y_true,
                     model_name, weight_rmse = 0.5, weights_mape = 0.0):
    
    # TODO -> funzione che riallinea le colonne di output di predizione (Id_pred)
    concat_df_test = concat_df_test.dropna()
    reconciled_df = reconciled_df.dropna()
    joined_df = reconciled_df.join(concat_df_test, how='inner')

    if len(joined_df) > 0:
        interest_columns = [col for col in reconciled_df.columns if col.startswith(f'{model_name}'.title()) and 'lo' not in col[-7:] and 'hi' not in col]
        
        metrics_methods = {}
        for pred_col in interest_columns:
            mape = round(mean_absolute_percentage_error(joined_df[y_true], joined_df[pred_col]), 2)
            rmse = round(mean_squared_error(joined_df[y_true], joined_df[pred_col], squared = False), 2)
            weighted_metric = weight_rmse * rmse + weights_mape * mape
            metrics_methods[pred_col] = weighted_metric

        # Check if the best reconciler is the base one
        # If that's the case, log a warning and calculate the second best reconciler
        best_reconciler = min(metrics_methods, key=lambda k: metrics_methods[k])
        if best_reconciler == f'{model_name}'.title():
            logger.warning('None of the reconcilers is better than the base model. Returns the second best reconciler')
            del metrics_methods[best_reconciler]
            interest_columns.pop(interest_columns.index(best_reconciler))
            best_reconciler = min(metrics_methods, key=lambda k: metrics_methods[k])
        # If that's not the case, return the proper index of the reconciler by removing the model_name
        else:
            del metrics_methods[f'{model_name}'.title()]
            interest_columns.pop(interest_columns.index(f'{model_name}'.title()))
        
        return interest_columns.index(best_reconciler), best_reconciler
    else:
        raise ValueError('There are no common indexes between the two dataframes')


from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self, Any
import pandas as pd
import pickle

from ..utility.tele_logger import logger
from ..utility.resources import *
from .datapreparation_utils import ModelPreprocess


class PreprocessingClass(BaseEstimator, TransformerMixin):

    def __init__(self,
                 usecase: str,
                 model: str,
                 ):
        """Pass all the parameters that can be set as explicit keyword
        arguments.
        
        Parameters
        ----------
        usecase: str
            Parameter identifying usecase into account
        model: str
            Parameter identifying model to use
            
        """

        self.usecase = usecase.lower()
        self.model  = model.lower()

        if self.usecase == "teleconsulto":
            self.preprocessing = ModelPreprocess()
        elif self.usecase == "televisita":
            pass
        elif self.usecase == "teleassistenza":
            pass
        elif self.usecase == "telemonitoraggio":
            pass
        return
    
    @log_exception(logger)
    def fit(self,
            X: pd.DataFrame,
            y: Any = None,
            **params: dict) -> Self:
        """Learn parameters useful for transforming data.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame.
        parmas: dict 
            specifies prameters to initialize preprocessing
        y : None
            Ignored (only for compatibility).
        """

        if self.usecase == "teleconsulto":
            if self.model == "prophet":
                 for key, value in params.items():
                     setattr(self, key, value)    
            pass
      
        return self


    def transform(self,
                X: pd.DataFrame) -> pd.DataFrame:
        """Preparing Data to feed models"

        Must be called only after calling fit method.
        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame
        time_granularity: str 
            specifies temporal granularity

        Returns
        -------
        pd.DataFrame 
            DataFrame identifying the timeseries 
        """

        if self.usecase == "teleconsulto":
            if self.model == "prophet":
                X = self.preprocessing.generate_time_series(X, 
                                                            self.date_col, 
                                                            self.target_col, 
                                                            self.time_granularity,
                                                            self.missing_data_strategy)
            else:
                pass
        elif self.usecase == "televisita":
            pass
        elif self.usecase == "teleassistenza":
            pass
        elif self.usecase == "telemonitoraggio":
            pass
        
        return X
    

    def save(self,
             path: str) -> None:
        """Store instance to file.
        Parameters
        ----------
        path : str
            Path to the file where the object must be stored.
        """

        with open(path, 'wb') as f:
            pickle.dump(self, f)
        
        return

    @classmethod
    def load(cls,
             path: str) -> Self:
        """Load instance from file.
        Parameters
        ----------
        path : str
            Path to the file where the object must be stored.
        """
        with open(path, 'rb') as f:
            return pickle.load(f)




import pandas as pd
import pytest
from tc_uc4.utility.experiment_utils import add_mapped_columns, generate_queries, create_zero_dataframe
from tc_uc4.utility.constants import code_to_region_name, code_to_speciality


@pytest.fixture
def sample_df():
    data = {
        'col1': ['A', 'A', 'B', 'B'],
        'col2': [30, 190, 50, 80],
        'Value': [1, 2, 3, 4]
    }
    return pd.DataFrame(data)


### add_mapped_columns ###

def test_add_mapped_columns(sample_df):
    hierarchy = {'Level1': 'col1', 'Level2': 'col2'}
    conversion = {'Level1': '', 'Level2': 'code_to_region_name'}
    expected_hierarchy = {'Level1': 'col1', 'Level2': 'col2_mapped'}
    expected_df = pd.DataFrame({
        'col1': ['A', 'A', 'B', 'B'],
        'col2': [30, 190, 50, 80],
        'Value': [1, 2, 3, 4],
        'col2_mapped': ['Lombardia', 'Sicilia', 'Veneto', 'Emilia-Romagna']
    })

    new_df, new_hierarchy = add_mapped_columns(hierarchy, sample_df, conversion)

    pd.testing.assert_frame_equal(new_df, expected_df)
    assert new_hierarchy == expected_hierarchy


### generate_queries ###

def test_generate_queries(sample_df):
    hierarchy_values = ['col1', 'col2']
    expected_dictionary = {
        'A': "(col1=='A')",
        'B': "(col1=='B')",
        'A/30': "(col1=='A' & col2==30)",
        'A/190': "(col1=='A' & col2==190)",
        'B/50': "(col1=='B' & col2==50)",
        'B/80': "(col1=='B' & col2==80)"
    }

    dictionary = generate_queries(hierarchy_values, sample_df)

    assert dictionary == expected_dictionary


### create_zero_dataframe ###

def test_create_zero_dataframe_valid_input():
    columns = ['A', 'B', 'C']
    n_rows = 3
    result = create_zero_dataframe(columns, n_rows)
    expected_result = pd.DataFrame({'A': {0: 0.0, 1: 0.0, 2: 0.0}, 
                                    'B': {0: 0.0, 1: 0.0, 2: 0.0}, 
                                    'C': {0: 0.0, 1: 0.0, 2: 0.0}})
    pd.testing.assert_frame_equal(result, expected_result)
    assert isinstance(result, pd.DataFrame)



import numpy as np
import pandas as pd
import os
import shutil

from .constants import code_to_region_name, code_to_speciality
from ..datapreparation.datapreparation_utils import logger

from ..algorithms.prophet_utils import train_val_test_split
from ..datapreparation.prep import PreprocessingClass
from ..datahandler.datahandler import DataHandler


def add_mapped_columns(
    hierarchy: dict[str, str], df: pd.DataFrame, conversion: dict[str, str]
) -> tuple[pd.DataFrame, dict[str, str]]:
    """Adds columns in the dataframe "df" according to a given hierarchy and conversion dictionary.

    Parameters
    ----------
    hierarchy : dict[str, str]
        A dictionary mapping levels to column names in the DataFrame.
    df : pd.DataFrame
        The input DataFrame to be mapped.
    conversion : dict[str, str]
        A dictionary mapping levels to conversion names.

    Returns
    -------
    tuple[pd.DataFrame, dict[str, str]]
        A tuple containing the modified DataFrame and the updated hierarchy dictionary.
    """
    for conversion_level in conversion.keys():
        if conversion[conversion_level]:
            name_column = hierarchy[conversion_level]
            dict_mapping = eval(conversion[conversion_level])
            df[name_column] = df[name_column].apply(
                lambda x: dict_mapping[x]
            )
            # hierarchy[conversion_level] = name_column
    return df #, hierarchy


def generate_queries(hierarchy_values: list[str], df: pd.DataFrame) -> dict[str, str]:
    """Generate a dictionary of ids and queries based on the given hierarchy values and DataFrame.

    Parameters
    ----------
    hierarchy_values : list[str]
        A list of hierarchy levels in the DataFrame.
    df : pd.DataFrame
        The DataFrame containing the data.


    Returns
    -------
    dict[str, str]
        Dictionary where the keys are unique identifiers and the values are queries.
    """
    dictionary = {}
    for n in range(1, len(hierarchy_values) + 1):
        df_aggregate = df.loc[:, hierarchy_values[:n]].drop_duplicates()
        for i in df_aggregate.itertuples(index=False):
            query = str(i)[6:].replace("=", "==").replace(",", " &")
            id_pred = "/".join([f"{i[j]}" for j in range(len(i))])
            dictionary[id_pred] = query
    return dictionary


def create_zero_dataframe(columns: list[str], n_rows: int) -> pd.DataFrame:
    """Creates a dataframe with the given columns and rows full of zeros.

    Parameters
    ----------
    columns : list[str]
        column names
    n_rows : int
        Number of rows

    Returns
    -------
    pd.DataFrame
        dataframe
    """
    # Create a dictionary with the column names and values
    data = {col: np.zeros(n_rows) for col in columns}

    # Create the dataframe
    df = pd.DataFrame(data)

    return df


def check_val_test_size(val_size: int, test_size: int):
    """Check the validation and test sizes to ensure they meet the minimum requirement.

    Parameters
    ----------
    val_size : int
        The size of the validation set.
    test_size : int
        The size of the test set.
    """
    if val_size == 0 or test_size == 0:
        logger.info(
            f"Be aware that either validation size {val_size} or test size {test_size} is equal to 0"
        )
    
    if test_size >= val_size:
         logger.error(f'Validation Set size MUST be greater than Test Size. Inputs given were: Validation Size -> {val_size} / Test Size -> {test_size}')
         raise ValueError(f'Validation Set size MUST be greater than Test Size. Inputs given were: Validation Size -> {val_size} / Test Size -> {test_size}')


def get_prophetexperiment_datasets(df: pd.DataFrame, dict_id_pred_queries: dict[str, str],
                                   preprocessor: PreprocessingClass,
                                   preprocessing_dict: dict
                                #    , time_granularity: str,
                                #    date_col: str
                                   ) -> dict:
    """
    _summary_

    Parameters
    ----------
    df
        _description_
    dict_id_pred_queries
        _description_
    preprocessor
        _description_
    preproc_details
        _description_

    Returns
    -------
        _description_
    """
    dict_preprocessed_df = {} 
    for id_pred, query in dict_id_pred_queries.items():
            filtered_df = df.query(query)
            filtered_df = filtered_df.reset_index(drop = True)
            filtered_df = allign_dates(filtered_df, preprocessing_dict['date_col'], df)
            full_df = preprocessor.fit_transform(filtered_df, **preprocessing_dict)
            dict_preprocessed_df[id_pred] = full_df
    return dict_preprocessed_df


def split_prophetexperiment_datasets(dict_preprocessed_df: dict, val_size: float, test_size: float = None) -> dict:
    """
    _summary_

    Parameters
    ----------
    dict_preprocessed_df
        _description_
    val_size
        _description_
    test_size, optional
        _description_, by default None

    Returns
    -------
        _description_
    """
    df_train_dict, df_val_dict, df_test_dict = {}, {}, {}
    for id_pred, dataset in dict_preprocessed_df.items():
        df_train, df_val, df_test = train_val_test_split(dataset, val_size, test_size)
        df_train_dict[id_pred] = df_train
        df_val_dict[id_pred] = df_val
        df_test_dict[id_pred] = df_test
    return {'train_dict': df_train_dict, 'val_dict' : df_val_dict, 'test_dict' : df_test_dict}


def allign_dates(filtered_df : pd.DataFrame, date_col: str, df: pd.DataFrame) -> pd.DataFrame:
    num = filtered_df.dtypes[filtered_df.dtypes.isin([int, float])].index
    obj = filtered_df.dtypes[filtered_df.dtypes == object].index
    max = df[date_col].max()
    min = df[date_col].min()
    if max not in filtered_df[date_col].values:
        new_line = len(filtered_df) # 0 based
        filtered_df.loc[new_line, num] = 0
        filtered_df.loc[new_line, obj] = filtered_df.loc[new_line-1, obj]
        filtered_df.loc[new_line, date_col] = max

    if min not in filtered_df[date_col].values:
        new_line = len(filtered_df) # 0 based
        filtered_df.loc[new_line, num] = 0
        filtered_df.loc[new_line, obj] = filtered_df.loc[new_line-1, obj]
        filtered_df.loc[new_line, date_col] = min
        filtered_df = filtered_df.sort_values(date_col).reset_index(drop = True)

    return filtered_df


def write_dataframes_locally(filenames_list, dataframes_list, save_data_folder):
            os.makedirs(save_data_folder, exist_ok=True)
            DH = DataHandler(data_folder = save_data_folder)
            for index in range(len(filenames_list)):
                DH.write(dataframes_list[index], filename = filenames_list[index], folder = '')
        
def remove_dataframe_locally(folder_to_remove):
     shutil.rmtree(folder_to_remove)


from __future__ import annotations

import os
import sys
from typing import Any, Dict

import tomli

def get_configuration(name: str, config_path : str, config_file: str = "web_app.toml") -> Dict[str, Any]:
    """Read the configurations set in the properties.toml file in the config folder and returns them
    as a dictionary

    Parameters
    ----------
    name : str
        Section to be 
    config_path : str
        Specifies the config path where the .toml configurations resides
    config_file : str, optional
        Specifies the config file to read, by default 'properties.toml'

    Returns
    -------
    dict
        Dictionary containing the properties read

    Raises
    ------
    Exception
        Raise error if .conf file is not found
    """

    if config_file not in os.listdir(config_path):
        raise Exception(
            f"File {config_file} not found in {config_path} please double check that the filename is correct."
        )

    config = tomli.load(open(os.path.join(config_path, config_file), "rb"))
    # Read from the specified file, the specified section
    if name in config:
        return config[name]
    else:
        raise Exception("System " + name + " not found in .conf file")

def log_exception(logger=None):
    """
    A decorator that wraps the passed in function and logs
    exceptions should one occur

    Parameters
    ----------
    logger : logging.Logger, default: None
        The logger to use, if None error and traceback will be printed anyway.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                # log the exception
                err = "There was an exception in "
                err += func.__name__
                print(err)
                if logger:
                    logger.exception(err)
                sys.exit()

        return wrapper

    return decorator


import pandas as pd
import pickle
import itertools
from sklearn.metrics import mean_squared_error, mean_absolute_error
import numpy as np
from prophet import Prophet
from typing_extensions import Self

from .prophet_utils import check_prophet_fit_df, check_prophet_predict_df, check_prophet_save, check_prophet_load
from ..utility.tele_logger import logger


class Prophet_model:
    def __init__(self, dic_param: dict = {
                  "changepoint_prior_scale": 0.5,
                  "seasonality_prior_scale": 0.1,
                  "yearly_seasonality": 10,
                  "weekly_seasonality": 3, 
                  "daily_seasonality": 10,
                  }) -> Self:
        """
        Istanciate the Prophet_model class

        Parameters
        ----------
        dic_param: dict        
            Dictionary representing the model parameters 

        Returns
        -------
        """
        logger.info('Initializing the model class', important = True)
        self.model = Prophet(**dic_param)

    def fit(self,
            df: pd.DataFrame) -> Self:
        """
        fit method: train the prophet model on df
        Parameters
        ----------
        df: pd.DataFrame
            DataFrame in the prophet format two columns: ['ds', 'y'])
        
        Returns
        -------
        """
        # The input dataset must be in the prophet format: it must contain the two columns ['ds', 'y']
        
        check_prophet_fit_df(df)
        self.df = df
        logger.info('Training the model', important = True)
        self.model.fit(self.df)
  
    def future_dataset(self,
                dic_param: dict = {'periods': 100,
                                   'freq': 'D'}):

        """
        Generation of the dataset for future forecasting

        Parameters
        ----------
        dic_param: dict
            Dictionary of the future parameters (e.g. 'periods', 'freq')

        Returns
        -------
            pd.DataFrame
            The dataframe for the forecasting in the future
        """
        logger.info('Creation of the future dataset for forecasting')
        return self.model.make_future_dataframe(**dic_param)
    
    def predict(self,
                df: pd.DataFrame) -> pd.DataFrame:
        """
        Predict method: prediction on df 

        Parameters
        ----------
        df: pd.DataFrame
        The dataframe on which the prediction is applied. 
            
        Returns
        -------
            pd.DataFrame
            The prophet dataframe for predictions
        """
        # The dataset should be either 2 columns ('ds','y') or 1 column ('ds')
        check_prophet_predict_df(df)

        logger.info('Prediction phase')
        return self.model.predict(df)
    
    def save(self,
             path: str) -> None:
        r"""Store algorithm to file.
        Parameters
        ----------
        path : str
            path of the file where the algorithm must be stored.
        """

        check_prophet_save(model = self.model, file_path = path)

        logger.info('Saving the model in the ".pkl" format', important = True)
        with open(path, 'wb') as file:
            pickle.dump(self.model, file)

    @classmethod
    def load(cls,
             path: str) -> Self:
        r"""Load algorithm from file.
        Parameters
        ----------
        path : str
            Path to the file where the algorithm must be stored.
        """
        check_prophet_load(file_path = path)

        logger.info('Loading the model', important = True)
        with open(path, 'rb') as file:
            model_file = pickle.load(file)
        loaded_model = cls({})
        loaded_model.model = model_file
        return loaded_model


def prophet_tuning(param_grid: dict, train_df: pd.DataFrame, validation_df: pd.DataFrame,
                   weight_rmse = 0.5, weight_mape = 0.5) -> dict:
    """
    Performs the parameters selection on a given set of parameters grid

    Parameters
    ----------
    param_grid: dict
        The parameters grid space

    train_df: pd.DataFrame
        The dataframe on which the model fit is applied 

    validation_df: pd.DataFrame
        The dataframe on which the model evaluation is applied 

    weight_rmse: numeric, optional
        weight of the rmse, by default 0.5

    weight_mape: numeric, optional
        weight of the mape, by default 0.5

    Returns
    -------
        dictionary of the best parameters

    Raises
    ------
    ValueError
        if the validation set is empty, after cleaning
    """
    rmse_mape = []
    keys, values = zip(*param_grid.items())
    permutations_dicts = [dict(zip(keys, v)) for v in itertools.product(*values)]
    for params in permutations_dicts:
        m =  Prophet_model(params)
        m.fit(train_df)
        validation_df_clean = validation_df.dropna()
        if len(validation_df_clean) > 0:
            val_predictions = m.model.predict(validation_df_clean)
            rmse = mean_squared_error(validation_df_clean['y'], val_predictions['yhat'], squared = False)
            mape = mean_absolute_error(validation_df_clean['y'], val_predictions['yhat'])
            rmse_mape.append(weight_rmse * rmse + weight_mape * mape)
        else:
            raise ValueError("Il dataframe di convalida con la rimozione dei valori nulli, è vuoto")

    best_params_index = np.argmin(rmse_mape)
    best_params = permutations_dicts[best_params_index]
    return best_params



import pytest
import pandas as pd
from tc_uc4.algorithms.algorithm import prophet_tuning, Prophet_model

### prophet_tuning ###

@pytest.fixture
def example_train_df():
    return pd.DataFrame({
        'ds': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'y': [10.5, 11.2, 9.8]
    })

@pytest.fixture
def example_validation_df():
    return pd.DataFrame({
        'ds': pd.to_datetime(['2023-06-04', '2023-06-05', '2023-06-06']),
        'y': [9.7, 11.0, 10.2]
    })

def test_prophet_tuning_valid(example_train_df, example_validation_df):
    param_grid = {
        'changepoint_prior_scale': [0.01],
        'seasonality_prior_scale': [0.5]
    }
    best_params = prophet_tuning(param_grid, example_train_df, example_validation_df)
    assert best_params == {'changepoint_prior_scale': 0.01, 'seasonality_prior_scale': 0.5}

def test_prophet_tuning_empty_validation_df(example_train_df):
    param_grid = {
        'changepoint_prior_scale': [0.01, 0.1, 0.5],
        'seasonality_prior_scale': [0.5, 5, 10]
    }
    empty_validation_df = pd.DataFrame(columns=['ds', 'y'])
    with pytest.raises(ValueError, match="Il dataframe di convalida con la rimozione dei valori nulli, è vuoto"):
        prophet_tuning(param_grid, example_train_df, empty_validation_df)


### Prophet_model ###

@pytest.fixture
def example_df_model():
    return pd.DataFrame({
        'ds': pd.to_datetime(['2023-06-01', '2023-06-02', '2023-06-03']),
        'y': [10.5, 11.2, 9.8]
    })

def test_prophet_model_fit(example_df_model):
    model = Prophet_model()
    model.fit(example_df_model)
    assert model.df.equals(example_df_model)

def test_prophet_model_predict(example_df_model):
    model = Prophet_model()
    model.fit(example_df_model)
    future_df = model.future_dataset()
    predictions = model.predict(future_df)
    assert isinstance(future_df, pd.DataFrame)
    assert isinstance(predictions, pd.DataFrame)

def test_prophet_model_save_load(tmp_path, example_df_model):
    model = Prophet_model()
    model.fit(example_df_model)

    # Save the model
    save_path = tmp_path / "model.pkl"
    model.save(str(save_path))

    # Load the model
    loaded_model = Prophet_model.load(str(save_path))

    assert isinstance(loaded_model, Prophet_model)
    assert loaded_model.model is not None
    pd.testing.assert_frame_equal(loaded_model.model.history, model.model.history)



from __future__ import annotations
from hierarchicalforecast.methods import BottomUp, MinTrace

code_to_region_name: dict[int, str] = {
                 30: 'Lombardia',190: 'Sicilia',50: 'Veneto',80: 'Emilia-Romagna',120: 'Lazio',200: 'Sardegna',
                 150: 'Campania',10: 'Piemonte',42: 'Trento',70: 'Liguria',130: 'Abruzzo',100: 'Umbria',
                 180: 'Calabria',160: 'Puglia',110: 'Marche',90: 'Toscana',20: "Valle D Aosta",170: 'Basilicata',
                 60: 'Friuli-Venezia-Giulia',41: 'Bolzano',140: 'Molise'}


code_to_speciality: dict[int, str] = {2: 'Cardiologia',
 3: 'Chirurgia Generale',
 13: 'Nefrologia',
 15: 'Neurologia',
 19: 'Ortopedia E Traumatologia',
 20: 'Ostetricia E Ginecologia',
 21: 'Otorinolaringoiatria',
 23: 'Psichiatria',
 12: 'Recupero E Riabilitazione',
 25: 'Urologia',
 18: 'Oncologia',
 16: 'Oculistica',
 5: 'Chirurgia Vascolare',
 27: 'Dermatologia',
 22: 'Pneumologia',
 10: 'Gastroenterologia',
 9: 'Endocrinologia'}


code_to_tipologia_medico: dict[str, str] = {"F" : "MMG",
    "P" : "PLS",
    "H" : "Ospedaliero",
    "A" : "Specialista ambulatoriale (ex SUMAI)",
    "G" : "Guardia medica",
    "T" : "Guardia medica turistica",
    "C" : "Specialista di struttura privata accreditata",
    "U" : "Medico di azienda ospedaliero-universitaria",
    "D" : "Dipendente dei servizi territoriali ASL",
    "I" : "Medico INAIL",
    "Z" : "Specializzazione non compresa tra le precedenti",
    "X" : "Altro (tirocinanti, specializzandi, etc.)"}


predict_dataframe_columns: list[str] = [
                        "Timestamp",
                        "Id_pred",
                        "Pred_mean",
                        "Sigma",
                        "Pi_lower_95",
                        "Pi_upper_95",
                    ]

format_ = '.parquet'
filenames_list = ["test_pred" + format_, "test_real" + format_,
                        "train_pred"+ format_, "train_real"+ format_,
                        "val_pred"+ format_, "val_real"+ format_]

reconcilers = [BottomUp(), MinTrace(method='ols')]


import pandas as pd
import pytest
from unittest.mock import MagicMock
from datetime import datetime
from typing import Any, Dict
from prophet import Prophet
from sklearn.utils.estimator_checks import check_transformer_general
from tc_uc4.algorithms.experiment_model import preprocess_and_split_df, PreprocessingClass, ExperimentModel, Prophet_model
from tc_uc4.datapreparation.prep import PreprocessingClass


@pytest.fixture
def sample_df():
    data = {
        'data_richiesta': pd.date_range(start='2022-01-01', end='2022-01-10'),
        'id_teleconsulto': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    }
    return pd.DataFrame(data)


### preprocess_and_split_df ###

def test_preprocess_and_split_df(sample_df):
    preprocessor = PreprocessingClass(usecase="teleconsulto", model="prophet")
    time_granularity = "D"
    val_size = 0.2
    test_size = 0.1

    # Test preprocess_and_split_df function
    result = preprocess_and_split_df(preprocessor, sample_df, time_granularity, val_size, test_size)

    # Check the returned tuple
    assert isinstance(result, tuple)
    assert len(result) == 4
    df_preproc, df_train, df_val, df_test = result

    # Check DataFrame types
    assert isinstance(df_preproc, pd.DataFrame)
    assert isinstance(df_train, pd.DataFrame)
    assert isinstance(df_val, pd.DataFrame)
    assert isinstance(df_test, pd.DataFrame)

    # Check DataFrame shapes
    assert df_preproc.shape == (10, 2)
    assert df_train.shape == (7, 2)
    assert df_val.shape == (2, 2)
    assert df_test.shape == (1, 2)


# def test_transformer_checks():
#     preprocessor = PreprocessingClass(usecase="teleconsulto", model="prophet")
#     check_transformer_general(preprocessor, enforce_estimator_tags=True)


### ExperimentModel ###

@pytest.fixture
def experiment_model():
    preprocessor = PreprocessingClass()
    return ExperimentModel(preprocessor)


def test_experiment_model(experiment_model):
    df = pd.DataFrame({
        'data_richiesta': pd.date_range(start='2023-01-01', end='2023-01-31'),
        'id_teleconsulto': [i for i in range(31)]
    })
    dict_id_pred_queries = {'id1': 'index>16', 'id2': 'index<15'}
    hyperparameters = {'weekly_seasonality':[True], 'yearly_seasonality':[True]}
    time_granularity = 'D'
    val_size = 0.2
    test_size = 0.1
    max_na_ratio = 0.5

    experiment_model.fit(
        df,
        dict_id_pred_queries,
        hyperparameters,
        time_granularity,
        val_size,
        test_size,
        max_na_ratio
    )

    assert len(experiment_model.models_dict) == 2 # Number of dataframes from dict_id_pred_queries

    # test create_hyperparameters_table

    result = experiment_model.create_hyperparameters_table(hyperparameters)

    assert len(result) == 2 # number of dataframes from dict_id_pred_queries
    assert 'id1' in result # key number 1 from dict_id_pred_queries
    assert 'id2' in result # key number 2 from dict_id_pred_queries
    # params ffrom models
    assert result['id1'] == {'weekly_seasonality':True, 'yearly_seasonality':True} # params for model from id1
    assert result['id2'] == {'weekly_seasonality':True, 'yearly_seasonality':True} # params for model from id2


    #test predict
    result = experiment_model.predict(
        df,
        dict_id_pred_queries,
        time_granularity,
        val_size,
        test_size
    )

    assert len(result) == 5 # len of the two test sets (3+2)
    assert result['Id_pred'].nunique() == 2 # number of dataframes (id1 and id2)

    # test evaluate

    result = experiment_model.evaluate(
        df,
        dict_id_pred_queries,
        time_granularity,
        val_size,
        test_size
    )

    assert len(result) == 2 # number of dataframes (id1 and id2)
    assert result.shape == (2,7)
    assert 'id1' in result['Id_pred'].unique()
    assert list(result.columns) == ['Id_pred', 'MAE', 'MAPE', 'RMSE', 'MSE', 'R2', 'Percentage Coverage']
    